package fr.ill.ics.n3d.io;

import java.io.File;
import java.util.logging.Logger;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import fr.ill.ics.n3d.link.Nomad3DController;
import fr.ill.ics.n3d.model.Axis;
import fr.ill.ics.n3d.model.Component;
import fr.ill.ics.n3d.model.ConfigParams;
import fr.ill.ics.n3d.model.Model;
import javafx.scene.paint.PhongMaterial;
import javafx.stage.FileChooser;
import javafx.stage.FileChooser.ExtensionFilter;

/**
 * Class for exporting Nomad 3D models.
 * @author dages
 */
public class Nomad3DExporter {

	/**
	 * Default constructor.
	 */
	public Nomad3DExporter() {
	}
	
	/**
	 * Exports the XML file of the model.
	 * @param model Model to export
	 * @param xmlPath Path of the XML file target
	 */
	public void write(Model model, String xmlPath) {
		try {
			Logger.getLogger("nomad-3d").info("Writing model \"" + model.getName() + "\" to \"" + xmlPath + "\"...\n");
			
			File outputFile = new File(xmlPath);

			DocumentBuilderFactory dbFactory = DocumentBuilderFactory.newInstance();
			DocumentBuilder dBuilder = dbFactory.newDocumentBuilder();
			Document doc = dBuilder.newDocument();
			
			// Create document
			Element eNomad3D = doc.createElement("Nomad3DXML");
			eNomad3D.setAttribute("state", model.getState().next().toString());
			
			Element eRoot = writeComponent(model.getRoot(), doc);
			eNomad3D.appendChild(eRoot);
			
			Element eGeometriesDirs = doc.createElement("GeometriesDirectories");
			String[] dirs = { // from the most detailed to the coarsest
					"exported",
					"decimated",
					"convex_hull"
			};
			for (int i = 0 ; i < dirs.length ; i++) {
				Element eDir = doc.createElement("Directory");
				eDir.setAttribute("lod", Integer.toString(i));
				eDir.setTextContent(dirs[i]);
				eGeometriesDirs.appendChild(eDir);
			}
			eNomad3D.appendChild(eGeometriesDirs);
			
			doc.appendChild(eNomad3D);
			doc.setXmlVersion("1.0");
			
			// Write xml file
			TransformerFactory tFactory= TransformerFactory.newInstance();
			Transformer transformer = tFactory.newTransformer();
			transformer.setOutputProperty(OutputKeys.INDENT, "yes");
			transformer.setOutputProperty("{http://xml.apache.org/xslt}indent-amount", "4");
			transformer.setOutputProperty(OutputKeys.STANDALONE, "yes");
			transformer.transform(new DOMSource(doc), new StreamResult(outputFile));
			
			Logger.getLogger("nomad-3d").info("Model \"" + model.getName() + "\" written.\n");
			
			// ZIP test
//			File zipFile = new File(outputFile.getParent() + File.separator + outputFile.getName().substring(0, outputFile.getName().length()-4) + ".n3d");
//			ZipUtil.pack(outputFile.getParentFile(), zipFile);
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	
	/**
	 * Exports the XML file of the model. The location of this file is provided by the user.
	 * @param model Model to export
	 */
	public void write(Model model) {
		FileChooser chooser = new FileChooser();
		chooser.setTitle("Save model");
		chooser.setInitialFileName(model.getName() + ".xml");
		chooser.getExtensionFilters().addAll(new ExtensionFilter("XML files", "*.xml"));
		File selectedFile = chooser.showSaveDialog(null);
		if (selectedFile != null) {
			write(model, selectedFile.getAbsolutePath());
		}
	}
	
	/**
	 * Writes the XML element of a component. The element is not added to the document.
	 * @param comp Component to write
	 * @param doc XML document
	 * @return XML element of the component
	 */
	private Element writeComponent(Component comp, Document doc) {
		String tagName = "Component";
		if (comp.isRoot()) {
			tagName = "Root" + tagName;
		}
		Element eComponent = doc.createElement(tagName);
		
		eComponent.setAttribute("name", comp.getName());
		eComponent.setAttribute("fileName", comp.getFileName());
		eComponent.setAttribute("wall", ((comp.isWall()) ? "True" : "False"));
		
		for (ConfigParams config : comp.getConfigurations()) {
			Element eConfig = writeConfigParams(config, doc);
			eComponent.appendChild(eConfig);
		}
		
		Element eMaterial = writeMaterial(comp.getMaterial(), doc);
		eComponent.appendChild(eMaterial);
		
		Element eAxis = writeAxis(comp.getAxis(), doc);
		eComponent.appendChild(eAxis);
		
		if (comp.getController() != null) {
			Element eController = writeController(comp.getController(), doc);
			eComponent.appendChild(eController);
		}
		
		for (Component child : comp.getChildren()) {
			Element eChild = writeComponent(child, doc);
			eComponent.appendChild(eChild);
		}
		
		return eComponent;
	}

	/**
	 * Writes the XML element of a configuration. The element is not added to the document.
	 * @param config Configuration to write
	 * @param doc XML document
	 * @return XML element for the configuration
	 */
	private Element writeConfigParams(ConfigParams config, Document doc) {
		Element eConfig = doc.createElement("ConfigParams");
		
		eConfig.setAttribute("configuration", config.getConfiguration());
		eConfig.setAttribute("visible", ((config.isVisible()) ? "True" : "False"));
		eConfig.setAttribute("fixed", ((config.isFixed()) ? "True" : "False"));
		
		Element eValue = doc.createElement("AxisValue");
		eValue.setTextContent(Double.toString(config.getAxisValue()));
		eConfig.appendChild(eValue);
		
		Element eTransform = doc.createElement("Transform");
		Element eRotation = doc.createElement("Rotation");
		eRotation.setTextContent(config.getRotation().getMxx()
				+ " " + config.getRotation().getMyx()
				+ " " + config.getRotation().getMzx()
				+ " " + config.getRotation().getMxy()
				+ " " + config.getRotation().getMyy()
				+ " " + config.getRotation().getMzy()
				+ " " + config.getRotation().getMxz()
				+ " " + config.getRotation().getMyz()
				+ " " + config.getRotation().getMzz());
		eTransform.appendChild(eRotation);
		Element eTranslation = doc.createElement("Translation");
		eTranslation.setTextContent(config.getTranslation().getTx()
				+ " " + config.getTranslation().getTy()
				+ " " + config.getTranslation().getTz());
		eTransform.appendChild(eTranslation);
		Element eScale = doc.createElement("Scale");
		eScale.setTextContent(Double.toString(config.getScale().getX()));
		eTransform.appendChild(eScale);
		eConfig.appendChild(eTransform);
		
		return eConfig;
	}
	
	/**
	 * Writes the XML element of a material. The element is not added to the document.
	 * @param material Material to write
	 * @param doc XML document
	 * @return XML element of the material
	 */
	private Element writeMaterial(PhongMaterial material, Document doc) {
		Element eMaterial = doc.createElement("Material");
		
		Element eDiffuse = doc.createElement("Diffuse");
		eDiffuse.setTextContent(material.getDiffuseColor().getRed() 
				+ " " + material.getDiffuseColor().getGreen() 
				+ " " + material.getDiffuseColor().getBlue() 
				+ " " + material.getDiffuseColor().getOpacity());
		eMaterial.appendChild(eDiffuse);
		
		Element eSpecular = doc.createElement("Specular");
		eSpecular.setTextContent(material.getSpecularColor().getRed() 
				+ " " + material.getSpecularColor().getGreen() 
				+ " " + material.getSpecularColor().getBlue() 
				+ " " + material.getSpecularColor().getOpacity());
		eMaterial.appendChild(eSpecular);
		
		Element eShininess = doc.createElement("Shininess");
		eShininess.setTextContent(Double.toString(material.getSpecularPower()));
		eMaterial.appendChild(eShininess);
		
		return eMaterial;
	}
	
	/**
	 * Writes the XML element of an axis. The element is not added to the document.
	 * @param axis Axis to write
	 * @param doc XML document
	 * @return XML element of the axis
	 */
	private Element writeAxis(Axis axis, Document doc) {
		Element eAxis = doc.createElement("Axis");
		
		eAxis.setAttribute("type", axis.getType().toString());
		eAxis.setAttribute("method", axis.getMethod().toString());
		
		Element eDir = doc.createElement("Direction");
		eDir.setTextContent(axis.getDirection().getX()
				+ " " + axis.getDirection().getY()
				+ " " + axis.getDirection().getZ());
		eAxis.appendChild(eDir);
		
		Element ePos = doc.createElement("Position");
		ePos.setTextContent(axis.getPosition().getX()
				+ " " + axis.getPosition().getY()
				+ " " + axis.getPosition().getZ());
		eAxis.appendChild(ePos);

		Element eZero = doc.createElement("ZeroValue");
		eZero.setTextContent(Double.toString(axis.getZeroValue()));
		eAxis.appendChild(eZero);
		
		Element eMin = doc.createElement("MinValue");
		eMin.setTextContent(Double.toString(axis.getMinValue()));
		eAxis.appendChild(eMin);
		
		Element eMax = doc.createElement("MaxValue");
		eMax.setTextContent(Double.toString(axis.getMaxValue()));
		eAxis.appendChild(eMax);
		
		return eAxis;
	}
	
	private Element writeController(Nomad3DController controller, Document doc) {
		Element eController = doc.createElement("Controller");
		
		eController.setAttribute("name", controller.getName());
		
		return eController;
	}

}
