package fr.ill.ics.n3d.io;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.util.Scanner;
import java.util.logging.Logger;

import javax.management.modelmbean.XMLParseException;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import fr.ill.ics.n3d.link.Nomad3DController;
import fr.ill.ics.n3d.link.NomadLinker;
import fr.ill.ics.n3d.model.AngleMate;
import fr.ill.ics.n3d.model.Axis;
import fr.ill.ics.n3d.model.CoincidentMate;
import fr.ill.ics.n3d.model.Component;
import fr.ill.ics.n3d.model.ConcentricMate;
import fr.ill.ics.n3d.model.ConfigParams;
import fr.ill.ics.n3d.model.DistanceMate;
import fr.ill.ics.n3d.model.Mate;
import fr.ill.ics.n3d.model.MateEntity;
import fr.ill.ics.n3d.model.Model;
import fr.ill.ics.n3d.model.Model.State;
import fr.ill.ics.n3d.model.ParallelMate;
import fr.ill.ics.n3d.model.PerpendicularMate;
import fr.ill.ics.n3d.model.TangentMate;
import javafx.geometry.Insets;
import javafx.geometry.Point3D;
import javafx.geometry.Pos;
import javafx.scene.Scene;
import javafx.scene.control.Label;
import javafx.scene.image.Image;
import javafx.scene.layout.StackPane;
import javafx.scene.paint.Color;
import javafx.scene.paint.PhongMaterial;
import javafx.scene.transform.Affine;
import javafx.scene.transform.Scale;
import javafx.stage.FileChooser;
import javafx.stage.FileChooser.ExtensionFilter;
import javafx.stage.Stage;
import javafx.stage.StageStyle;

/**
 * Class for importing Nomad 3D models. 
 * 
 * The models are stored in one XML file and a set of STL files. The STL files contain the geometries of the parts and 
 * the XML file contain all of the other informations.
 * 
 * @author dages
 */
public class Nomad3DImporter {
		
	/** Imported model. */
	private Model model;
	
	private String currentComponentName;
	
	/**
	 * Default constructor.
	 */
	public Nomad3DImporter() {
		this.model = new Model();
	}
	
	/**
	 * Imports an image from the jar resource folder.
	 * @param imageName Path of the image, relative to the resource folder
	 * @return The image if the import is successful, null otherwise
	 */
	public static Image createImage(String imageName) {
        InputStream stream = Nomad3DImporter.class.getClassLoader().getResourceAsStream(imageName);
        return (stream != null) ? new Image(stream) : null;
    }
	
	/**
	 * Gets the model. 
	 * @return The model
	 */
	public Model getModel() {
		return this.model;
	}
	
	/**
	 * Reads a Nomad 3D model.
	 * @param xmlPath Path of the XML file describing the model
	 */
	public void read(String xmlPath) {
		
		Logger.getLogger("nomad-3d").info("Reading model " + xmlPath + "...\n");
		this.model.clear();
		try {
			File inputFile = new File(xmlPath);
			if (!inputFile.exists()) {
				throw new IOException("File '" + xmlPath + "' does not exist");
			}
			String dirPath = inputFile.toPath().toAbsolutePath().getParent().toString();

			DocumentBuilderFactory dbFactory = DocumentBuilderFactory.newInstance();
			DocumentBuilder dBuilder = dbFactory.newDocumentBuilder();
			Document doc = dBuilder.parse(inputFile);
			doc.getDocumentElement().normalize();

			if (doc.getDocumentElement().getNodeName() != "Nomad3DXML") {
				throw new XMLParseException("Wrong root name while parsing " + xmlPath + " (" + doc.getDocumentElement().getNodeName() + ")");
			}

			readModel(doc, dirPath);
			this.model.setName(inputFile.getName().substring(0, inputFile.getName().length()-4));
			this.model.setDirectoryPath(dirPath);
			if (model.getState() == Model.State.EXPORTED) {
				readMates(doc);
			} else {
				this.model.computeSceneHierarchies();
			}

			Logger.getLogger("nomad-3d").info("Model " + xmlPath + " read.\n");
			printMemoryUse();
		} catch (Exception e) {
			e.printStackTrace();
			Logger.getLogger("nomad-3d").severe(e.getMessage() + "\n");
		}
	}

	/**
	 * Reads a Nomad 3D model. A dialog box is used to select the xml file.
	 */
	public void read() {
		FileChooser chooser = new FileChooser();
		chooser.setTitle("Open model");
		chooser.getExtensionFilters().addAll(new ExtensionFilter("XML files", "*.xml"));
		File selectedFile = chooser.showOpenDialog(null);
		if (selectedFile != null) {
			Stage readingStage = new Stage(StageStyle.UTILITY);
			readingStage.setTitle("Open model");
			Label readingLabel = new Label("Reading model " + selectedFile.getName() + "...");
			readingLabel.setPrefHeight(50);
			readingLabel.setPrefWidth(10*selectedFile.getName().length());
			readingLabel.setAlignment(Pos.CENTER);
			StackPane readingPane = new StackPane(readingLabel);
			readingPane.setAlignment(Pos.CENTER);
			StackPane.setMargin(readingPane, new Insets(10));
			Scene readingScene = new Scene(readingPane);
			readingStage.setScene(readingScene);
			readingStage.show();
			read(selectedFile.getAbsolutePath());
			readingStage.close();
		}
	}
	
	/**
	 * Reads alll of the configurations.
	 * @param doc XML document
	 * @param dirPath Path of the directory containing the files
	 * @throws XMLParseException if the document is invalid
	 */
	private void readModel(Document doc, String dirPath) throws XMLParseException {
		Element eNomad3D = (Element)doc.getElementsByTagName("Nomad3DXML").item(0);
		Model.State state = Model.State.fromString(eNomad3D.getAttribute("state"));
		this.model.setState(state);
		Element eRoot = (Element)eNomad3D.getElementsByTagName("RootComponent").item(0);
		Component root = readComponent(eRoot, dirPath);
		this.model.setRoot(root);
		if (state != State.EXPORTED) {
			readGeometriesDirectories(eNomad3D);
		}
	}
		
	/**
	 * Reads a component and its sub-hierarchy.
	 * @param eComponent Component's XML element
	 * @param dirPath Path of the directory containing the files
	 * @return Component read
	 * @throws XMLParseException if the document is invalid
	 */
	private Component readComponent(Element eComponent, String dirPath) throws XMLParseException {

		String name = eComponent.getAttribute("name");
		currentComponentName = name;
		
		String fileName = eComponent.getAttribute("fileName");
		String wall = eComponent.getAttribute("wall");
		
		Component component = new Component();
		component.setName(name);
		component.setFileName(fileName);
		if (wall != null) {
			component.setWall(wall.equals("True"));
		}
		
		AsciiStringConverter asciiConverter = new AsciiStringConverter();
		if (name.contains("&")) {
			System.out.println("ASCII converter : " + name + " --> " + asciiConverter.convert(name));
			System.out.println("ASCII converter : " + fileName + " --> " + asciiConverter.convert(fileName));
		}
		
		readComponentConfigurations(eComponent, component);
		
		Element eMaterial = (Element)eComponent.getElementsByTagName("Material").item(0);
		PhongMaterial material = readMaterial(eMaterial);
		component.setMaterial(material);
		
		readAxis(eComponent, component);
		
		readController(eComponent, component);
		
		NodeList children = eComponent.getChildNodes();
		for (int i = 0 ; i < children.getLength() ; i++) {
			Node nChild = children.item(i);
			if (nChild.getNodeType() == Node.ELEMENT_NODE && nChild.getNodeName().equals("Component")) {
				Component child = readComponent((Element)nChild, dirPath);
				if (child != null) {
					component.addChild(child);
				}
			}
		}

		if (model.getState() == State.EXPORTED) {
			if (component.isLeaf()) {
				File stlFile = new File(dirPath, fileName + ".STL");
				if (!stlFile.exists()) {
					return null;
				}
			}
		}
		
		return component;
	}
	
	/**
	 * Reads the configurations of a component.
	 * @param eComponent Component's XML element
	 * @param component Component owning the configurations
	 */
	private void readComponentConfigurations(Element eComponent, Component component) {

		NodeList children = eComponent.getChildNodes();
		for (int i = 0 ; i < children.getLength() ; i++) {
			Node child = children.item(i);
			if (child.getNodeType() != Node.ELEMENT_NODE) {
				continue;
			}
			if (child.getNodeName().equals("ConfigParams")) {
				ConfigParams curConfig = readConfigParams((Element)child, component);
				component.addConfiguration(curConfig);
			}
		}
	}
	
	/**
	 * Reads a configuration of a component
	 * @param eConfig Configuration's XML element
	 * @param owner Component owning the configuration
	 * @return The configuration
	 */
	private ConfigParams readConfigParams(Element eConfig, Component owner) {
		ConfigParams config = new ConfigParams(owner);
		
		String configName = eConfig.getAttribute("configuration");
		config.setConfiguration(configName);
		boolean fixed = eConfig.getAttribute("fixed").equals("True");
		config.setFixed(fixed);
		boolean visible = eConfig.getAttribute("visible").equals("True");
		config.setVisible(visible);
		
		// Axis value
		Element eAxisValue = (Element)eConfig.getElementsByTagName("AxisValue").item(0);
		double axisValue = Double.parseDouble(eAxisValue.getTextContent());
		config.setAxisValue(axisValue);
		
		// Transform
		Element eTransform = (Element)eConfig.getElementsByTagName("Transform").item(0);
		readTransform(eTransform, config.getRotation(), config.getTranslation(), config.getScale());
		
		// The mate group will be read in a second pass
		
		return config;
	}
	
	/**
	 * Reads the transform of an element.
	 * @param transformElement Transform's XML element
	 * @param rotation [out] Rotation part of the transform 
	 * @param translation [out] Translation part of the transform 
	 * @param scale [out] Scale part of the transform 
	 */
	private void readTransform(Element transformElement, Affine rotation, Affine translation, Scale scale) {
		// Order of the matrix elements :
		// http://help.solidworks.com/2015/English/api/sldworksapi/SolidWorks.Interop.sldworks~SolidWorks.Interop.sldworks.IMathTransform.html?verRedirect=1
		Scanner scanner = null;
		
		// Scale inversion in the blender script, maybe we should do it here ?
		Node nScale = transformElement.getElementsByTagName("Scale").item(0);
		scanner = new Scanner(nScale.getTextContent());
		double scaleFactor = scanner.nextDouble();
		scale.setX(scaleFactor);
		scale.setY(scaleFactor);
		scale.setZ(scaleFactor);
		scanner.close();

		Node nRotation = transformElement.getElementsByTagName("Rotation").item(0);
		scanner = new Scanner(nRotation.getTextContent());
		rotation.setMxx(scanner.nextDouble());
		rotation.setMyx(scanner.nextDouble());
		rotation.setMzx(scanner.nextDouble());
		rotation.setMxy(scanner.nextDouble());
		rotation.setMyy(scanner.nextDouble());
		rotation.setMzy(scanner.nextDouble());
		rotation.setMxz(scanner.nextDouble());
		rotation.setMyz(scanner.nextDouble());
		rotation.setMzz(scanner.nextDouble());
		scanner.close();

		Node nTranslation = transformElement.getElementsByTagName("Translation").item(0);
		scanner = new Scanner(nTranslation.getTextContent());
		translation.setTx(scanner.nextDouble());
		translation.setTy(scanner.nextDouble());
		translation.setTz(scanner.nextDouble());
		scanner.close();
	}
	
	/**
	 * Reads the material of a component.
	 * @param eMaterial Material's XML element
	 * @return The imported material
	 */
	private PhongMaterial readMaterial(Element eMaterial) {
		Scanner scanner = null;
		String defaultAttr = eMaterial.getAttribute("default");
		if (defaultAttr.equals("True")) {
			return Component.DEFAULT_MATERIAL;
		} else {
			try {
				Node nDiffuse = eMaterial.getElementsByTagName("Diffuse").item(0);
				scanner = new Scanner(nDiffuse.getTextContent());
				Color diffuse = Color.color(scanner.nextDouble(), scanner.nextDouble(), scanner.nextDouble(), scanner.nextDouble());
				scanner.close();
				Node nSpecular = eMaterial.getElementsByTagName("Specular").item(0);
				scanner = new Scanner(nSpecular.getTextContent());
				Color specular = Color.color(scanner.nextDouble(), scanner.nextDouble(), scanner.nextDouble(), scanner.nextDouble()); 
				scanner.close();
				Node nShininess = eMaterial.getElementsByTagName("Shininess").item(0); 
				scanner = new Scanner(nShininess.getTextContent());
				double shininess = scanner.nextDouble(); 
				scanner.close();
				
				PhongMaterial material = new PhongMaterial();
				material.setDiffuseColor(diffuse);
				material.setSpecularColor(specular);
				material.setSpecularPower(shininess);
				return material;
			}
			catch (Exception e) {
				System.err.println("Error while parsing Material element of the component '" + currentComponentName + "'");
				throw e;
			}
		}
	}
	
	/**
	 * Reads the axis of a component.
	 * @param eComponent Component's XML element
	 * @param component Component
	 */
	private void readAxis(Element eComponent, Component component) {
		Scanner scanner = null;
		Element eAxis = null;
		NodeList children = eComponent.getChildNodes();
		for (int i = 0 ; i < children.getLength() ; i++) {
			Node child = children.item(i);
			if (child.getNodeType() == Node.ELEMENT_NODE && child.getNodeName().equals("Axis")) {
				eAxis = (Element)child;
				break;
			}
		}
		if (eAxis == null) {
			Logger.getLogger("nomad-3d").severe("No axis found for " + component.getName());
			return;
		}
		
		String typeStr = eAxis.getAttribute("type");
		String methodStr = eAxis.getAttribute("method");		
		Axis.Type type = Axis.Type.fromString(typeStr);
		Axis.Method method = Axis.Method.fromString(methodStr);
		
		Axis axis = new Axis();
		axis.setType(type);
		axis.setMethod(method);

		Node nDirection = eAxis.getElementsByTagName("Direction").item(0);
		scanner = new Scanner(nDirection.getTextContent());
		Point3D direction = new Point3D(scanner.nextDouble(), scanner.nextDouble(), scanner.nextDouble());
		scanner.close();
		Node nPosition = eAxis.getElementsByTagName("Position").item(0);
		scanner = new Scanner(nPosition.getTextContent());
		Point3D position = new Point3D(scanner.nextDouble(), scanner.nextDouble(), scanner.nextDouble());
		scanner.close();
		
		Node nZeroValue = eAxis.getElementsByTagName("ZeroValue").item(0);
		if (nZeroValue != null) {
			double zeroValue = Double.parseDouble(nZeroValue.getTextContent());
			axis.setZeroValue(zeroValue);
		}
		
		Node nMinValue = eAxis.getElementsByTagName("MinValue").item(0);
		if (nMinValue != null) {
			double minValue = Double.parseDouble(nMinValue.getTextContent());
			axis.setMinValue(minValue);
		}
		
		Node nMaxValue = eAxis.getElementsByTagName("MaxValue").item(0);
		if (nMaxValue != null) {
			double maxValue = Double.parseDouble(nMaxValue.getTextContent());
			axis.setMaxValue(maxValue);
		}
		
		axis.setDirection(direction);
		axis.setPosition(position);
		
		component.setAxis(axis);
	}
	
	/**
	 * Reads the controller of a component, if there is one
	 * @param eComponent XML element of the component
	 * @param component Nomad 3D component
	 */
	private void readController(Element eComponent, Component component) {
		Element eController = null;
		NodeList children = eComponent.getChildNodes();
		for (int i = 0 ; i < children.getLength() ; i++) {
			Node child = children.item(i);
			if (child.getNodeType() == Node.ELEMENT_NODE && child.getNodeName().equals("Controller")) {
				eController = (Element)child;
				break;
			}
		}
		if (eController != null) {
			String controllerName = eController.getAttribute("name");
			Nomad3DController controller = NomadLinker.getInstance().getControllerByName(controllerName);
			if (controller != null && controller.getComponent() == null) {
				controller.setComponent(component);
				component.setController(controller);
			} else if (controller != null) {
				Logger.getLogger("nomad-3d").severe("Reading controller of " + component.getName() 
						+ " : controller already associated with another component (" 
						+ controller.getComponent().getName() + ")");
			} else {
				Logger.getLogger("nomad-3d").severe("Reading controller of " + component.getName() 
						+ " : the controller " + controllerName + " does not exist");
			}
		} 
	}
	
	/**
	 * Reads all the mates of the model.
	 * @param doc XML document describing the model
	 */
	private void readMates(Document doc) {
		Element eRoot = (Element)doc.getElementsByTagName("RootComponent").item(0);
		Component root = this.model.getRoot();
		readComponentMates(eRoot, root);
	}
	
	/**
	 * Reads mate groups for every components in the subhierarchy of the specified component. 
	 * The hierarchy must be read before.
	 * @param eComponent Component's XML element
	 * @param component Component Owner of the mates
	 */
	private void readComponentMates(Element eComponent, Component component) {
		NodeList children = eComponent.getChildNodes();
		
		// Config params of this component
		for (int i = 0 ; i < children.getLength() ; i++) {
			Node nChild = children.item(i);
			if (nChild.getNodeType() == Node.ELEMENT_NODE && nChild.getNodeName().equals("ConfigParams")) {
				Element eConfig = (Element)children.item(i);
				String configName = eConfig.getAttribute("configuration");
				ConfigParams config = component.getConfigurationByName(configName);
				if (config == null) {
					Logger.getLogger("nomad-3d").severe("Unable to find configuration " + configName + " in component " + component.getName());
				} else {
					readConfigMates(eConfig, config, component);
				}
			}
		}	
		
		// Children 
		int nbXmlChildren = 0;
		for (int i = 0 ; i < children.getLength() ; i++) {
			Node nChild = children.item(i);
			if (nChild.getNodeType() == Node.ELEMENT_NODE && nChild.getNodeName().equals("Component")) {
				Element eChild = (Element)children.item(i);
				String childName = eChild.getAttribute("name");
				Component child = component.getChildByName(childName, false);
				if (child == null) {
					Logger.getLogger("nomad-3d").warning("Unable to find child " + childName + " for component " + component.getName());
				} else {
					readComponentMates(eChild, child);
				}
				nbXmlChildren++;
			}
		}
		if (component.getChildren().size() != nbXmlChildren) {
			Logger.getLogger("nomad-3d").warning("XML element's number of children (" + children.getLength() + ") different than component's number of children (" + component.getChildren().size() + ") in " + component.getName());
		}
	}
	
	/**
	 * Reads the mate group of a configuration.
	 * @param eConfig Configuration's XML element
	 * @param config Configuration
	 * @param owner Component owning the configuration
	 */
	private void readConfigMates(Element eConfig, ConfigParams config, Component owner) {
		NodeList matesList = eConfig.getElementsByTagName("MateGroup");
		if (matesList.getLength() > 0) {
			Element eMates = (Element)matesList.item(0);
			NodeList mateList = eMates.getElementsByTagName("Mate");
			for (int i = 0 ; i < mateList.getLength() ; i++) {
				readMate((Element)mateList.item(i), config, owner);
			}
		}
	}
	
	/**
	 * Reads a Mate XML element.
	 * @param eMate Mate's XML element
	 * @param config Configuration owning the mate group 
	 * @param owner Owner of the configuration
	 */
	private void readMate(Element eMate, ConfigParams config, Component owner) {
		Scanner scanner = null;
		String name = eMate.getAttribute("name");
		Mate.Alignment alignment = Mate.Alignment.fromString(eMate.getAttribute("alignment"));
		double min = Double.parseDouble(eMate.getAttribute("min"));
		double max = Double.parseDouble(eMate.getAttribute("max"));
		boolean flipped = eMate.getAttribute("flipped").equals("True");
		
		Mate mate = null;
		String type = eMate.getAttribute("type");
		switch (type) {
			case CoincidentMate.XML_TYPE_NAME:
				mate = new CoincidentMate();
				break;
			case ConcentricMate.XML_TYPE_NAME:
				mate = new ConcentricMate();
				break;
			case DistanceMate.XML_TYPE_NAME:
				mate = new DistanceMate();
				break;
			case ParallelMate.XML_TYPE_NAME:
				mate = new ParallelMate();
				break;
			case PerpendicularMate.XML_TYPE_NAME:
				mate = new PerpendicularMate();
				break;
			case TangentMate.XML_TYPE_NAME:
				mate = new TangentMate();
				break;
			case AngleMate.XML_TYPE_NAME:
				mate = new AngleMate();
				break;
			default:
				Logger.getLogger("nomad-3d").severe("Unknown mate type : \"" + type +  "\", ignoring mate " + name);
				return;
		}
		mate.setName(name);
		mate.setAlignment(alignment);
		mate.setMin(min);
		mate.setMax(max);
		mate.setFlipped(flipped);
		
		NodeList entityList = eMate.getElementsByTagName("Entity");
		
		if (entityList.getLength() != Mate.NB_ENTITIES) {
			Logger.getLogger("nomad-3d").severe("Wrong number of entities : " + entityList.getLength() + " instead of " + Mate.NB_ENTITIES);
		}
		for (int i = 0 ; i < entityList.getLength() ; i++) {
			Element eEntity = (Element)entityList.item(i);
			
			MateEntity.Type entityType = MateEntity.Type.fromString(eEntity.getAttribute("type"));
			if (entityType == MateEntity.Type.UNSUPPORTED) {
				return;
			}
			
			Node eComponentPath = eEntity.getElementsByTagName("ComponentPath").item(0);
			String strPath = eComponentPath.getTextContent();
			String[] path = strPath.split("/");
			if (path.length == 0) { 
				Logger.getLogger("nomad-3d").severe("Target path is empty (mate " + name + ")");
			} else if (!owner.getName().equals(path[0])) {
				Logger.getLogger("nomad-3d").severe("Target path do not begin with mate group's owner \n" 
						+ "\t Path : " + strPath + "\n\t Owner : " + owner.getName());
			}
			Component target = owner;
			for (int j = 1 ; j < path.length ; j++) {
				target = target.getChildByName(path[j], false);
				if (target == null) {
					String msg = "Target not found, mate ignored \n";
					msg += "\t target : " + strPath + "\n";
					msg += "\t name : " + name + "\n\t type : " + type + "\n";
					msg += "\t mate group's owner : " + owner.getName() + "\n";
					Logger.getLogger("nomad-3d").severe(msg);
					return;
				}
			}
			
			Node nPosition = eEntity.getElementsByTagName("Position").item(0);
			scanner = new Scanner(nPosition.getTextContent());
			Point3D position = new Point3D(scanner.nextDouble(), scanner.nextDouble(), scanner.nextDouble());
			scanner.close();
			Node nAxis = eEntity.getElementsByTagName("Axis").item(0);
			scanner = new Scanner(nAxis.getTextContent());
			Point3D axis = new Point3D(scanner.nextDouble(), scanner.nextDouble(), scanner.nextDouble());
			scanner.close();
			Node nRadius = eEntity.getElementsByTagName("Radius").item(0);
			scanner = new Scanner(nRadius.getTextContent());
			double[] radius = new double[MateEntity.NB_RADIUS];
			for (int j = 0 ; j < MateEntity.NB_RADIUS ; j++) {
				radius[j] = scanner.nextDouble();
			}
			scanner.close();
			
			MateEntity entity = new MateEntity();
			entity.setTarget(target);
			entity.setType(entityType);
			entity.setAxis(axis);
			entity.setPosition(position);
			for (int j = 0 ; j < MateEntity.NB_RADIUS ; j++) {
				entity.setRadius(j, radius[j]);
			}
			
			mate.setEntity(i, entity);
		}
		
		config.addMate(mate);
	}
	
	/**
	 * Reads the geometries directories and store them in the model.
	 * @param eNomad3D XML element of the model
	 */
	private void readGeometriesDirectories(Element eNomad3D) {
		Element eGeomDirs = (Element)eNomad3D.getElementsByTagName("GeometriesDirectories").item(0);
		NodeList dirList = eGeomDirs.getChildNodes();
		for (int i = 0 ; i < dirList.getLength() ; i++) {
			Node nDir = dirList.item(i);
			if (nDir.getNodeType() == Node.ELEMENT_NODE && nDir.getNodeName().equals("Directory")) {
				Element eDir = (Element)nDir;
				Integer lod = Integer.parseInt(eDir.getAttribute("lod"));
				String dirName = eDir.getTextContent();
				this.model.addGeometryDirectory(lod, dirName);
			}
		}
	}
	
	/**
	 * Prints the memory usage of the application. For debug purpose.
	 */
	private static void printMemoryUse() {
		int mb = 1024*1024;
		Runtime runtime = Runtime.getRuntime();
		String msg = "Heap utilization : \n";
		msg += "Used Memory : " 
				+ (runtime.totalMemory() - runtime.freeMemory()) / mb
				+ " MB \n";
		msg += "Free Memory : " 
				+ runtime.freeMemory() / mb
				+ " MB \n";
		msg += "Total Memory : " + runtime.totalMemory() / mb
				+ " MB \n";
		msg += "Max Memory : " + runtime.maxMemory() / mb
				+ " MB \n";
		Logger.getLogger("nomad-3d").info(msg);
	}

}
