package fr.ill.ics.n3d.link;

import java.util.ArrayList;
import java.util.logging.Logger;

import fr.ill.ics.bridge.Controller;
import fr.ill.ics.core.property.Property;
import fr.ill.ics.core.property.PropertyManager;
import fr.ill.ics.n3d.model.Component;

/**
 * Wrapper for Nomad controller
 * @author dages
 */
public class Nomad3DController {
	
	/**
	 * Type of the axis of a controller.
	 * @author dages
	 */
	public enum Type {
		/** Rotation axis. */
		ROTATION {
			@Override
			public void init() {
				this.unitValues = new ArrayList<String>();
				this.unitValues.add("");
				this.unitValues.add("deg");
				this.unitValues.add("degree");
			}
		},
		/** Translation axis. */
		TRANSLATION {
			@Override
			public void init() {
				this.unitValues = new ArrayList<String>();
				this.unitValues.add("mm");
				this.unitValues.add("cm");
			}
		},
		/** Step controller, unknown axis type. */
		STEP {
			@Override
			public void init() {
				this.unitValues = new ArrayList<String>();
				this.unitValues.add("step");
			}
		};
		
		/** Possible representations of the unit of the type. */
		protected ArrayList<String> unitValues;
		
		/**
		 * Default constructor.
		 */
		private Type() {
			init();
		}
		
		/**
		 * Initialization of the type's unit representations.
		 */
		protected abstract void init();
		
		/**
		 * Conversion of a string to a type.
		 * @param unitValue Unit representation 
		 * @return The type corresponding to the unit if it exists, null otherwise
		 */
		public static Type fromUnitValue(String unitValue) {
			if (ROTATION.unitValues.contains(unitValue)) {
				return Type.ROTATION;
			} else if (TRANSLATION.unitValues.contains(unitValue)) {
				return Type.TRANSLATION;
			} else if (STEP.unitValues.contains(unitValue)) {
				return Type.STEP;
			} else {
				Logger.getLogger("nomad-3d").severe("Unknown unit value \"" + unitValue + "\"");
				return null;
			}
		}
		
		/**
		 * String representation of the type. 
		 * @return First unit representation
		 */
		@Override
		public String toString() {
			return this.unitValues.get(0);
		}
	}

	/** Wrapped controller. */
	private Controller controller;
	
	/** Attached component. */
	private Component component;
	
	/**
	 * Constructor.
	 * @param controller Controller to wrap
	 */
	public Nomad3DController(Controller controller) {
		this.controller = controller;
		this.component = null;
	}
	
	/**
	 * Gets the associated component.
	 * @return Component if it is set, null otherwise
	 */
	public Component getComponent() {
		return component;
	}

	/**
	 * Gets the name of the controller.
	 * @return The name
	 */
	public String getName() {
		return controller.getName();
	}
	
	/**
	 * Gets the value of a property.
	 * @param propertyName Property name
	 * @return Value of the property
	 */
	public String getPropertyValue(String propertyName) {
		Property prop = PropertyManager.getInstance().getProperty(controller, propertyName);
		return prop.getValue();
	}

	/**
	 * Gets the server value of a property.
	 * @param propertyName Property name
	 * @return Server value of the property
	 */
	public String getPropertyServerValue(String propertyName) {
		Property prop = PropertyManager.getInstance().getProperty(controller, propertyName);
		return prop.getServerValue();
	}
	
	/**
	 * Gets the server value of the actual position.
	 * @return The actual position
	 */
	public double getActualPosition() {
		switch (this.getName()) {
			case "DAN":
				return Double.parseDouble(getPropertyServerValue("actual_angle"));
			case "DH":
				return Double.parseDouble(getPropertyServerValue("actual_height"));
			default:
				return Double.parseDouble(getPropertyServerValue("actual_position"));
		}
	}

	/**
	 * Gets the server value of the wanted position.
	 * @return The wanted position
	 */
	public double getWantedPosition() {
		switch (this.getName()) {
			case "DAN":
				return Double.parseDouble(getPropertyServerValue("wanted_angle"));
			case "DH":
				return Double.parseDouble(getPropertyServerValue("wanted_height"));
			default:
				return Double.parseDouble(getPropertyServerValue("wanted_position"));
		}
	}

	/**
	 * Gets the server value of the offset position.
	 * @return The offset position
	 */
	public double getOffsetPosition() {
		return Double.parseDouble(getPropertyServerValue("offset_position"));
	}

	/**
	 * Gets the server value of the min position.
	 * @return The min position
	 */
	public double getMinPosition() {
		return Double.parseDouble(getPropertyServerValue("min_position"));
	}

	/**
	 * Gets the server value of the max position.
	 * @return The max position
	 */
	public double getMaxPosition() {
		return Double.parseDouble(getPropertyServerValue("max_position"));
	}

	/**
	 * Gets the type of the controller, based on the server value of the unit.
	 * @return The type
	 */
	public Type getType() {
		return Type.fromUnitValue(getPropertyServerValue("unit"));
	}
	
	/**
	 * Sets the associated component.
	 * @param component New component
	 */
	public void setComponent(Component component) {
		this.component = component;
	}
	
	/**
	 * String representation of the controller.
	 * @return Name of the controller
	 */
	@Override
	public String toString() {
		return this.getName();
	}
}
