package fr.ill.ics.n3d.link;

import java.util.Collection;
import java.util.HashMap;
import java.util.Set;
import java.util.logging.Logger;

import fr.ill.ics.bridge.ControllerManager;
import fr.ill.ics.bridge.LoginManager;
import fr.ill.ics.bridge.LoginManager.ClientAlreadyLaunchedException;
import fr.ill.ics.bridge.LoginManager.ConnectionFailure;
import fr.ill.ics.bridge.LoginManager.LoginIncorrectException;
import fr.ill.ics.bridge.command.CommandZoneWrapper;
import fr.ill.ics.cameo.Application;
import fr.ill.ics.core.property.PropertyManager;
import fr.ill.ics.nscclient.servant.ControllerServant;
import fr.ill.ics.nscclient.serverconnection.ServerInstance;
import fr.ill.ics.util.ConfigManager;
import javafx.application.Platform;

/**
 * Class handling the link between the editor and the server.
 * @author dages
 */
public class NomadLinker {
	public final static int LOGIN_OK = 0; 
	public final static int LOGIN_NOT_OK = 1; 
	public final static int LOGIN_REFUSED = 2;

	/** Instance of the linker. */
	private final static NomadLinker instance = new NomadLinker();
	
	/** Controllers loaded from Nomad server. */
	private HashMap<String, Nomad3DController> controllers;
	
	/**
	 * Default constructor.
	 */
	private NomadLinker() {
		controllers = new HashMap<>();
	}

	/**
	 * Gets the instance.
	 * @return The instance
	 */
	public static NomadLinker getInstance() {
		return instance;
	}
	
	/**
	 * Initializes the connection to cameo and nomad server.
	 * @param args Command line arguments of the program
	 */
	public void init(String[] args) {
		Logger.getLogger("nomad-3d").info("Initializing nomad server connection...");
		
		Application.This.init(args);

		// Initialize the ConfigManager at the very beginning because it is in this class where all .nomad/settings.properties are initialized!
		ConfigManager.initInstance();
		
		ServerInstance.getInstance().init();
		
		Logger.getLogger("nomad-3d").info("Nomad server endpoint : " + ConfigManager.getInstance().getNomadServerEndpoint());
		
		int status = login("", "", true);
		Logger.getLogger("nomad-3d").info("Nomad server login status : " + status);

		PropertyManager.getInstance().init();
		
		// Real axes
		Set<String> controllersNames = ControllerManager.getInstance().getControllersOfType("axis", true);
		
		// Virtual axes
		controllersNames.addAll(ControllerManager.getInstance().getControllersOfType("det_angle", true));
		controllersNames.addAll(ControllerManager.getInstance().getControllersOfType("det_height", true));
		controllersNames.addAll(ControllerManager.getInstance().getControllersOfType("san_coupled", true));
		
		for (String controllerName : controllersNames) {
			ControllerServant controller = ControllerManager.getInstance().getController(controllerName);
			controllers.put(controllerName, new Nomad3DController(controller));
		}

		Logger.getLogger("nomad-3d").info("Nomad server connection initialized.");
	}
	
	/**
	 * Gets the names of the controllers.
	 * @return Set of controllers names
	 */
	public Set<String> getControllersNames() {
		return controllers.keySet();
	}
	
	/**
	 * Gets the controllers.
	 * @return The controllers
	 */
	public Collection<Nomad3DController> getControllers() {
		return this.controllers.values();
	}
	
	/**
	 * Gets a specific controller.
	 * @param name Name of the controller
	 * @return Specified controller
	 */
	public Nomad3DController getControllerByName(String name) {
		return this.controllers.get(name);
	}
	
	/**
	 * Closes the connection to cameo and nomad server.
	 */
	public void terminate() {
		Logger.getLogger("nomad-3d").info("Terminating nomad server connection...");
		
		controllers.clear();
		
		// Close connection with Nomad Server
		LoginManager.getInstance().logout(CommandZoneWrapper.SERVER_ID);
		
		Application.This.terminate();
		
		Logger.getLogger("nomad-3d").info("Nomad server connection terminated.");
	}
	
	/**
	 * Log in nomad server.
	 * @param login Login
	 * @param password Password
	 * @param standalone 
	 * @return Login status
	 */
    private int login(String login, String password, boolean standalone) {
    	int loginState = LOGIN_NOT_OK;
        // Check login and password
        try {
            LoginManager.getInstance().login(login, password, standalone, CommandZoneWrapper.SERVER_ID);
            loginState = LOGIN_OK;
        } catch (LoginIncorrectException e) {
        	loginState = LOGIN_NOT_OK;
		} catch (ClientAlreadyLaunchedException e) {
			loginState = LOGIN_REFUSED;
		} catch (ConnectionFailure e) {
			Platform.exit();
		}
        
        return loginState;
    }
}
