/**
 * 
 */
package fr.ill.ics.n3d.model;

import java.util.HashSet;

import javafx.geometry.Point2D;
import javafx.geometry.Point3D;
import javafx.scene.DepthTest;
import javafx.scene.Group;
import javafx.scene.paint.Color;
import javafx.scene.paint.PhongMaterial;
import javafx.scene.shape.Box;
import javafx.scene.shape.Cylinder;
import javafx.scene.transform.Affine;
import javafx.scene.transform.Rotate;
import javafx.scene.transform.Transform;
import javafx.scene.transform.Translate;

/**
 * Class representing the axis of a component. 
 * @author dages
 */
public class Axis {
	
	/**
	 * Allowed axis types.
	 */
	public enum Type {
		/** 1 translational DOF. */
		TRANSLATION("Translation", Color.color(1, 1, 0)),
		/** 1 rotational DOF. */
		ROTATION("Rotation", Color.color(1, 0, 1)),
		/** No DOF. */
		FIXED("Fixed", Color.color(0, 0, 0)),
		/** 6 DOFs, ie the component is free. */
		NONE("None", Color.color(1, 1, 1));
		
		/** String representation of the type. */
		private String name;
		
		/** Color associated with the type. Used for visualizing axes. */
		private Color color;
		
		/**
		 * Constructor.
		 * @param name String representation of the type
		 * @param color Color associated with the type
		 */
		Type(String name, Color color) {
			this.name = name;
			this.color = color;
		}
		
		/**
		 * Conversion of a string to an axis type.
		 * @param name String representation of the type
		 * @return The type if the name is known, null otherwise
		 */
		public static Type fromString(String name) {
			switch(name) {
				case "Translation":
					return TRANSLATION;
				case "Rotation":
					return ROTATION;
				case "Fixed":
					return FIXED;
				case "None":
					return NONE;
				default:
					System.err.println("Axis.Type.fromString : ERROR : unknown type \"" + name + "\"");
					return null;
			}
		}
		
		/**
		 * Conversion of this type to a string.
		 * @return String representation of the type
		 */
		@Override
		public String toString() {
			return this.name;
		}
	}
	
	/**
	 * Methods used to compute an axis.
	 * @author dages
	 */
	public enum Method {
		/** Axis computed by this application. */
		COMPUTED("Computed"),
		/** Axis added by the user. */
		ADDED("Added"),
		/** No info, computation needed. */
		TO_COMPUTE("ToCompute");
		
		/** String representation of the method. */
		private String name;
		
		/**
		 * Constructor.
		 * @param name String representation of the method
		 */
		Method(String name) {
			this.name = name;
		}

		/**
		 * Conversion of a string to an axis method.
		 * @param name String representation of the method
		 * @return The method if the name is known, null otherwise
		 */
		public static Method fromString(String name) {
			switch(name) {
				case "Computed":
					return COMPUTED;
				case "Added":
					return ADDED;
				case "ToCompute":
					return TO_COMPUTE;
				default:
					return null;
			}
		}

		/**
		 * Conversion of this method to a string.
		 * @return String representation of the method
		 */
		@Override
		public String toString() {
			return this.name;
		}
	}
	
	/** Free axis, 6 DOFs. */
	public static final Axis FREE_AXIS;
	static {
		FREE_AXIS = new Axis();
		FREE_AXIS.setType(Type.NONE);
		FREE_AXIS.setMethod(Method.COMPUTED);
	}
	
	/** Fixed axis, 0 DOFs. */
	public static final Axis FIXED_AXIS;
	static {
		FIXED_AXIS = new Axis();
		FIXED_AXIS.setType(Type.FIXED);
		FIXED_AXIS.setMethod(Method.COMPUTED);
	}
	
	/** Threshold, used for axes computations. */
	private static final double EPSILON = 1e-1;
	
	/** Type of the axis. */
	private Type type;
	
	/** Method used to get this axis. */
	private Method method;
	
	/** Direction of the axis. This should be a unit vector. */
	private Point3D direction;
	
	/** Position of the axis, in scene coordinate system. */
	private Point3D position;
	
	/** Current value of the axis. */
	private double value;
	
	/** Minimum value of the axis. */
	private double minValue;
	
	/** Maximum value of the axis. */
	private double maxValue;
	
	/** 
	 * Zero value of the axis, set while calibrating.
	 */
	private double zeroValue;
	
	/** Transform applied by the user when moving the component along its axis. */
	private Affine movementTransform;
	
	/** Visual group of the axis. */
	private Group visualGroup;
	
	/** 
	 * Transform used convert scene coordinates into parent's coordinates. 
	 * This parent is either the parent of the axis' component or the component itself if it is the root.
	 */
	private Transform sceneToParent;

	/**
	 * Default constructor.
	 */
	public Axis() {
		this.type = Type.NONE;
		this.method = Method.TO_COMPUTE;
		this.direction = new Point3D(0, 0, 0);
		this.position = new Point3D(0, 0, 0);
		this.value = 0;
		this.minValue = Double.NEGATIVE_INFINITY;
		this.maxValue = Double.POSITIVE_INFINITY;
		this.zeroValue = 0;
		this.movementTransform = new Affine();
		this.visualGroup = new Group();
		this.visualGroup.setVisible(false);
		this.sceneToParent = new Affine();
	}
	
	/**
	 * Copy constructor.
	 * @param a Axis to copy
	 */
	public Axis(Axis a) {
		this.type = a.type;
		this.method = a.method;
		this.direction = new Point3D(a.direction.getX(), a.direction.getY(), a.direction.getZ());
		this.position = new Point3D(a.position.getX(), a.position.getY(), a.position.getZ());
		this.value = a.value;
		this.minValue = a.minValue;
		this.maxValue = a.maxValue;
		this.zeroValue = a.zeroValue;
		this.movementTransform = a.movementTransform.clone();
		this.visualGroup = new Group();
		this.visualGroup.setVisible(false);
		this.sceneToParent = a.sceneToParent.clone();
		updateVisualGroup();
	}
	
	/**
	 * Gets the type of the axis
	 * @return The type
	 */
	public Type getType() {
		return this.type;
	}

	/**
	 * Gets the method of the axis
	 * @return The method
	 */
	public Method getMethod() {
		return this.method;
	}

	/**
	 * Gets the direction of the axis
	 * @return The direction
	 */
	public Point3D getDirection() {
		return this.direction;
	}

	/**
	 * Gets the position of the axis
	 * @return The position
	 */
	public Point3D getPosition() {
		return this.position;
	}

	/**
	 * Gets the value of the axis
	 * @return The value
	 */
	public double getValue() {
		return this.value;
	}

	/**
	 * Gets the minimum value of the axis
	 * @return The minimum value
	 */
	public double getMinValue() {
		return minValue;
	}

	/**
	 * Gets the maximum value of the axis
	 * @return The maximum value
	 */
	public double getMaxValue() {
		return maxValue;
	}

	/**
	 * Gets the zero value of the axis, set while calibrating the axis.
	 * @return The zero value
	 */
	public double getZeroValue() {
		return zeroValue;
	}

	/**
	 * Gets the movement transform of the axis
	 * @return The movement transform
	 */
	public Affine getMovementTransform() {
		return movementTransform;
	}

	/**
	 * Gets the visual group of the axis
	 * @return The visual group
	 */
	public Group getVisualGroup() {
		return this.visualGroup;
	}

	/**
	 * Gets the scene to parent transform of the axis
	 * @return The scene to parent transform
	 */
	public Transform getSceneToParent() {
		return sceneToParent;
	}
	
	/**
	 * Sets the type of the axis. 
	 * @param type New type
	 */
	public void setType(Type type) {
		this.type = type;
		updateVisualGroup();
	}
	
	/**
	 * Sets the method of the axis
	 * @param method New method
	 */
	public void setMethod(Method method) {
		this.method = method;
	}
	
	/**
	 * Sets the direction of the axis. The new direction should be a unit vector. No verification is performed.
	 * @param direction New direction. 
	 */
	public void setDirection(Point3D direction) {
		this.direction = direction;
		updateVisualGroup();
	}
	
	/**
	 * Sets the position of the axis.
	 * @param position New position
	 */
	public void setPosition(Point3D position) {
		this.position = position;
		updateVisualGroup();
	}
	
	/**
	 * Sets the value of the axis, without moving it.
	 * @see #move(double)
	 * @param value New value
	 */
	public void setValue(double value) {
		this.value = value;
	}
	
	/**
	 * Sets the minimum value of the axis.
	 * @param minValue New minimum value
	 */
	public void setMinValue(double minValue) {
		this.minValue = minValue;
	}
	
	/**
	 * Sets the maximum value of the axis.
	 * @param maxValue New maximum value
	 */
	public void setMaxValue(double maxValue) {
		this.maxValue = maxValue;
	}
	
	/**
	 * Sets the zero value of the axis.
	 * @param zeroValue New zero value
	 */
	public void setZeroValue(double zeroValue) {
		this.zeroValue = zeroValue;
	}
	
	/**
	 * Sets the scene to parent transform of the axis.
	 * @param sceneToParent New transform
	 */
	public void setSceneToParent(Transform sceneToParent) {
		this.sceneToParent = sceneToParent;
		updateVisualGroup();
	}
	
	/**
	 * Updates the visual group of the axis. 
	 * 
	 * This method should be called when any of the following attributes of  the axis is changed :
	 * <ul>
	 * <li> direction </li>
	 * <li> position </li>
	 * <li> type </li>
	 * <li> scene to parent transform </li>
	 * </ul>
	 */
	private void updateVisualGroup() {
		this.visualGroup.getChildren().clear();
		this.visualGroup.getTransforms().clear();
		
		PhongMaterial material = new PhongMaterial();
		material.setDiffuseColor(this.type.color);
		material.setSpecularColor(this.type.color);
		if (!(this.type == Type.NONE || this.type == Type.FIXED)) {
			Box box = new Box(2, 2, 100);
			box.setMaterial(material);
			box.getTransforms().add(new Translate(0, 0, 50));
			this.visualGroup.getChildren().add(box);
			
			Cylinder cylinder = new Cylinder(3, 10);
			cylinder.setMaterial(material);
			cylinder.getTransforms().addAll(new Translate(0, 0, 100), new Rotate(90, Rotate.X_AXIS));
			this.visualGroup.getChildren().add(cylinder);
			
			Point3D zAxis = Rotate.Z_AXIS;
			Translate translation = new Translate(this.position.getX(), this.position.getY(), this.position.getZ());
			this.visualGroup.getTransforms().add(translation);
			Rotate rotation = new Rotate();
			if (zAxis.crossProduct(this.direction).magnitude() > EPSILON) {
				rotation.setAxis(zAxis.crossProduct(this.direction));
				rotation.setAngle(zAxis.angle(this.direction));
			} else if (zAxis.dotProduct(this.direction) < 0) {
				rotation.setAxis(Rotate.X_AXIS);
				rotation.setAngle(180);
			}
			this.visualGroup.getTransforms().add(rotation);
		}
		this.visualGroup.getTransforms().add(0, this.sceneToParent);
		this.visualGroup.setDepthTest(DepthTest.DISABLE);
	}

	/**
	 * Computes the axis described by a set of SolidWorks mates (constraints).
	 * @param mates Set of mates 
	 * @param config Configuration owning the mates
	 * @return The computed axis
	 */
	public static Axis fromMates(HashSet<Mate> mates, ConfigParams config) {
		Axis translation = new Axis();
		translation.setType(Type.NONE);
		translation.setMethod(Method.COMPUTED);
		translation.setDirection(Point3D.ZERO);
		translation.setPosition(Point3D.ZERO);
		Axis rotation = new Axis();
		rotation.setType(Type.NONE);
		rotation.setMethod(Method.COMPUTED);
		rotation.setDirection(Point3D.ZERO);
		rotation.setPosition(Point3D.ZERO);
		
		// Entities attributes are in the assembly coordinate system
		// Transform rootToScene = root.getTranslation().createConcatenation(root.getRotation().createConcatenation(root.getScale()));
		Transform rootToScene = config.getTranslation().createConcatenation(config.getRotation().createConcatenation(config.getScale()));
		
		for (Mate mate : mates) {
			mate.apply(translation, rotation, rootToScene);
			
			if (translation.getType() == Type.FIXED && rotation.getType() == Type.FIXED) {
				break;
			}
		}
		
		return (rotation.getType() == Type.ROTATION) ? rotation : translation; // priority to rotation
		// return (translation.getType() == Type.TRANSLATION) ? translation : rotation; / priority to translation
	}
	
	public void moveTo(double value) {
		this.move(value - this.value);
	}
	
	/**
	 * Moves the axis according to a value. After this call, the value of the axis is incremented by the moving value.
	 * @param value Moving value
	 */
	public void move(double value) {
		if (Math.abs(value) < EPSILON) {
			return;
		}
		double coef = value;
		if (this.value + coef > this.maxValue) {
			coef = this.maxValue - this.value;
		} else if (this.value + coef < this.minValue) {
			coef = this.minValue - this.value;
		}
		this.value += coef;
		
		switch (this.type) {
			case TRANSLATION:
				this.movementTransform.appendTranslation(
						coef * this.direction.getX(), 
						coef * this.direction.getY(), 
						coef * this.direction.getZ()
					);
				break;
			case ROTATION:
				if (Math.abs(coef) > EPSILON) {
					this.movementTransform.appendRotation(
							coef, 
							this.position, 
							this.direction
						);
				}
				break;
			default:
				break;
		}
	}
	
	/**
	 * Moves the axis. This is a convenience method to handle mouse entries.
	 * @param delta Mouse movement's delta
	 * @param screenAxisDir Axis direction in screen coordinates
	 * @param multiplier Multiplier of the movement amplitude
	 */
	void move(Point2D delta, Point2D screenAxisDir, double multiplier) {
		double coef = 0;
		switch (this.type) {
			case TRANSLATION:
				coef = delta.dotProduct(screenAxisDir);
				break;
			case ROTATION:
				coef = screenAxisDir.angle(delta);
				while (coef < 0) coef += 360;
				while (coef > 180) coef -= 180;
				coef = (coef/90) - 1;
				if (Math.abs(coef) < 1) {
					coef = Math.exp(-1/(1 - coef*coef));
				} else {
					coef = 0;
				}
				coef *= delta.magnitude() * -Math.signum(delta.dotProduct(screenAxisDir.getY(), -screenAxisDir.getX()));
				break;
			default:
				break;
		}
		this.move(multiplier * coef);
	}
	
	/**
	 * Resets the movement to the initial position.
	 */
	public void resetMovement() {
		this.value = 0;
		this.movementTransform.setToIdentity();
		this.move(0);
	}
	
	/**
	 * Resets the movement.
	 */
	public void setToZeroMovement() {
		this.value = zeroValue;
		this.movementTransform.setToIdentity();
		this.move(zeroValue);
	}
	
	/**
	 * Clones this axis.
	 * @return A copy of this axis
	 */
	@Override
	public Axis clone() {
		return new Axis(this);
	}

	@Override
	public String toString() {
		String str = "Axis [\n";
		str += "Type : " + this.type + "\n";
		str += "Method : " + this.method + "\n";
		str += "Direction : " + this.direction + "\n";
		str += "Position : " + this.position + "\n";
		str += "Value : " + this.value + "\n";
		str += "Min value : " + this.minValue + "\n";
		str += "Max value : " + this.maxValue + "\n";
		str += "]";
		return str;
	}
}
