package fr.ill.ics.n3d.model;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import java.util.Stack;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.concurrent.CopyOnWriteArraySet;
import java.util.logging.Logger;

import com.interactivemesh.jfx.importer.stl.StlMeshImporter;

import fr.ill.ics.n3d.link.Nomad3DController;
import javafx.beans.value.ChangeListener;
import javafx.beans.value.ObservableValue;
import javafx.event.EventHandler;
import javafx.geometry.Point2D;
import javafx.geometry.Point3D;
import javafx.scene.Group;
import javafx.scene.Node;
import javafx.scene.control.CheckBoxTreeItem;
import javafx.scene.input.MouseEvent;
import javafx.scene.paint.Color;
import javafx.scene.paint.PhongMaterial;
import javafx.scene.shape.CullFace;
import javafx.scene.shape.MeshView;
import javafx.scene.shape.TriangleMesh;
import javafx.scene.shape.VertexFormat;
import javafx.scene.transform.Affine;
import javafx.scene.transform.Rotate;
import javafx.scene.transform.Transform;

/**
 * Class describing a component.
 * @author dages
 */
public class Component {
	
	/** Default material. */
	public static final PhongMaterial DEFAULT_MATERIAL;
	static {
		DEFAULT_MATERIAL = new PhongMaterial(Color.WHITE);
		DEFAULT_MATERIAL.setSpecularColor(Color.color(0, 0, 0, 0));
		DEFAULT_MATERIAL.setSpecularPower(250);
	}
	/** Flag used to turn on the cursor selection. */
	public static boolean selectOn = false;
	/** Flag used to turn on the cursor unselection. */
	public static boolean unselectOn = false;
	
	/** Component currently moved by the user. Null if nothing moves. */
	private static Component movingComponent = null;
	/** Current mouse x position. */
	private static double mousePosX = 0;
	/** Current mouse y position. */
	private static double mousePosY = 0;
	/** Previous mouse x position. */
	private static double mouseOldX = 0;
	/** Previous mouse y position. */
	private static double mouseOldY = 0;
	
	/** Current index for inserting components. Used in the axes computation. */
	private static int insertedIndex = 1;
	/** Threshold */
	private static final double EPSILON = 1e-1;
	
	/** Name of the component. */
	private String name;
	/** File name of the component. Basically the name without the SW instance index. */
	private String fileName;
	
	/** Scene node of the component; */
	private Node sceneNode;
	
	/** Item in the tree view */
	private CheckBoxTreeItem<Component> treeItem;
	
	/** Material of the component. */
	private PhongMaterial material;
	
	private CopyOnWriteArrayList<ConfigParams> configurations;
		
	/** Children of the component. */
	private CopyOnWriteArrayList<Component> children;
	/** parent of the component. Null if it is the root.*/
	private Component parent;
	
	/** Axis of the component. */
	private Axis axis;
	
	/** Either the component is a wall or not. */
	private boolean wall;
	
	/** Initial axis of the component. Used internally. */
	private Axis initialAxis;
	/** 
	 * List of the axes computed for each configuration of the component. 
	 * These are used internally for axes computation. 
	 */
	private ArrayList<Axis> computedAxes;
	
	/** Attached controller, describing the Nomad axis. */
	private Nomad3DController controller;
	
	/**
	 * Default constructor.
	 */
	public Component() {
		this.name = "";
		this.fileName = "";
//		this.directoryPath = "";
		this.material = DEFAULT_MATERIAL;
		this.sceneNode = null;
		this.children = new CopyOnWriteArrayList<Component>();
		this.parent = null;
		this.axis = null;
		this.initialAxis = null;
		this.computedAxes = new ArrayList<>();
		this.configurations = new CopyOnWriteArrayList<>();
		this.treeItem = null;
		this.wall = false;
		this.controller = null;
	}
	
	/**
	 * Gets the name of the component. 
	 * 
	 * This name is made of the file name of the component and an instance number, as follows : filename-1
	 * 
	 * @return The name
	 */
	public String getName() {
		return this.name;
	}
	
	/**
	 * Gets the file name of the component.
	 * @return The file name
	 */
	public String getFileName() {
		return this.fileName;
	}
	
	/**
	 * gets the material of the component.
	 * @return The material 
	 */
	public PhongMaterial getMaterial() {
		return this.material;
	}
	
	/**
	 * Gets the scene node of the component.
	 * 
	 * If the components is a leaf this node is a MeshView, otherwise it is a Group.
	 * 
	 * @return The scene node.
	 */
	public Node getSceneNode() {
		return this.sceneNode;
	}
	
	/**
	 * Gets the configurations of the component.
	 * @return Array of the configurations
	 */
	public List<ConfigParams> getConfigurations() {
		return configurations;
	}
	
	/**
	 * Gets the children of the component.
	 * @return Array of the children
	 */
	public CopyOnWriteArrayList<Component> getChildren() {
		return this.children;
	}
	
	/**
	 * Gets the parent of the component.
	 * 
	 * If the component is the root, its parent is null.
	 * 
	 * @return The parent
	 */
	public Component getParent() {
		return this.parent;
	}
	
	/**
	 * Gets the axis of the component.
	 * @return The axis
	 */
	public Axis getAxis() {
		return this.axis;
	}
	
	/**
	 * Gets the wall flag of the components
	 * @return true if the component is a wall
	 */
	public boolean isWall() {
		return wall;
	}
	
	/**
	 * Get the tree item of the component.
	 * @return The tree item
	 */
	public CheckBoxTreeItem<Component> getTreeItem() {
		return treeItem;
	}
	
	/**
	 * Gets the Nomad controller.
	 * @return The controller
	 */
	public Nomad3DController getController() {
		return controller;
	}
	
	/**
	 * Sets the name of the component.
	 * @param name New name
	 */
	public void setName(String name) {
		this.name = name;
	}
	
	/**
	 * Sets the file name of the component.
	 * @param fileName New file name
	 */
	public void setFileName(String fileName) {
		this.fileName = fileName;
	}
	
	/**
	 * Sets the material of the component and its descendants.
	 * @param material New material
	 * @param recurse true to set the material of all the sub-hierarchy, false to change this component only
	 */
	public void setMaterial(PhongMaterial material, boolean recurse) {
		this.material = material;
		if (this.sceneNode != null && this.isLeaf()) {
			MeshView meshView = (MeshView)this.sceneNode;
			meshView.setMaterial(this.material);
		}
		if (recurse) {
			for (Component child : this.children) {
				child.setMaterial(material, recurse);
			}
		}
	}
	
	/**
	 * Sets the material of the component.
	 * @param material New material
	 */
	public void setMaterial(PhongMaterial material) {
		setMaterial(material, false);
	}
	
	/**
	 * Sets the scene node of the component.
	 * @param node New node
	 */
	public void setSceneNode(Node node) {
		this.sceneNode = node;
	}
	
	/**
	 * Sets the axis of the component.
	 * @param axis New axis
	 */
	public void setAxis(Axis axis) {
		this.axis = axis;
		if (this.initialAxis == null) {
			this.initialAxis = this.axis;
		}
		resetAxisTransform(false);
	}
	
	/**
	 * Marks the component as a wall.
	 * @param wall New wall flag
	 */
	public void setWall(boolean wall) {
		this.wall = wall;
	}
	
	/**
	 * Sets the Nomad controller
	 * @param controller New controller
	 */
	public void setController(Nomad3DController controller) {
		this.controller = controller;
	}

	/**
	 * Adds a child to this component. If the child already has a parent, it is removed from its former parent children.
	 * @param child New child
	 */
	public void addChild(Component child) {
		if (getChildByName(child.getName()) != null) {
			System.err.println("Component.addChild : Duplicating " + child.getName() + " in " + this.name);
		}
		if (child.parent != null) {
			child.parent.children.remove(child);
		}
		this.children.add(child);
		child.parent = this;
	}
	
	/**
	 * Removes a child, if it is one.
	 * @param child Child to remove
	 * @return true if the child was a child of this component
	 */
	public boolean removeChild(Component child) {
		if (child == null || !child.isChild(this)) {
			return false;
		}
		child.parent = null;
		if (this.sceneNode != null) {
			try {
				((Group)this.sceneNode).getChildren().remove(child.getSceneNode());
				((Group)this.sceneNode).getChildren().remove(child.getAxis().getVisualGroup());
			} catch (Exception e) {
				e.printStackTrace();
			}
		}
		return this.children.remove(child);
	}
	
	/**
	 * Adds a configuration to this component.
	 * @param config New configuration
	 */
	public void addConfiguration(ConfigParams config) {
		removeConfiguration(config.getConfiguration(), false);
		this.configurations.add(config);
	}
	
	public void removeConfiguration(String configName, boolean recursive) {
		for (ConfigParams config : this.configurations) {
			if (config.getConfiguration().equals(configName)) {
				this.configurations.remove(config);
			}
		}
		if (recursive) {
			for (Component child : this.children) {
				child.removeConfiguration(configName, recursive);
			}
		}
	}
	
	/**
	 * Saves the current position of the component and its descendants in a new configuration.
	 * @param configName Name of the new configuration
	 */
	public void saveCurrentConfiguration(String configName, String configSourceName, boolean recursive) {
		ConfigParams newConfig = new ConfigParams(this);
		newConfig.setConfiguration(configName);
		newConfig.setTransform(this.getSceneNode().getLocalToSceneTransform());
		newConfig.setAxisValue(this.axis.getValue());

		ConfigParams configSource = getConfigurationByName(configSourceName);
		if (configSource != null) {
			newConfig.setVisible(configSource.isVisible());
		}
		
		addConfiguration(newConfig);
		
		if (recursive) {
			for (Component child : this.children) {
				child.saveCurrentConfiguration(configName, configSourceName, recursive);
			}
		}
	}
	
	public void saveCurrentConfiguration(String configName, String configSourceName) {
		this.saveCurrentConfiguration(configName, configSourceName, true);
	}
	
	public void saveMedianConfiguration(String configSourceName) {
		double oldAxisValue = this.axis.getValue();
		
		this.axis.moveTo(this.axis.getZeroValue());
		saveCurrentConfiguration("median", configSourceName, false);
		
		for (Component child : this.children) {
			child.saveMedianConfiguration(configSourceName);
		}
		
		this.axis.moveTo(oldAxisValue);
	}
	
	public void saveMinConfiguration(String configSourceName) {
		double oldAxisValue = this.axis.getValue();

		if (!Double.isInfinite(this.axis.getMinValue()) && !Double.isNaN(this.axis.getMinValue())) {
			this.axis.moveTo(this.axis.getMinValue());
		} else {
			this.axis.moveTo(this.axis.getZeroValue());
		}
		saveCurrentConfiguration("min", configSourceName, false);
		if (Double.isInfinite(this.axis.getMinValue()) || Double.isNaN(this.axis.getMinValue())) {
			this.getConfigurationByName("min").setAxisValue(this.axis.getMinValue());
		}
		
		for (Component child : this.children) {
			child.saveMinConfiguration(configSourceName);
		}
		
		this.axis.moveTo(oldAxisValue);
	}
	
	public void saveMaxConfiguration(String configSourceName) {
		double oldAxisValue = this.axis.getValue();

		if (!Double.isInfinite(this.axis.getMaxValue()) && !Double.isNaN(this.axis.getMaxValue())) {
			this.axis.moveTo(this.axis.getMaxValue());
		} else {
			this.axis.moveTo(this.axis.getZeroValue());
		}
		saveCurrentConfiguration("max", configSourceName, false);
		if (Double.isInfinite(this.axis.getMaxValue()) || Double.isNaN(this.axis.getMaxValue())) {
			this.getConfigurationByName("max").setAxisValue(this.axis.getMaxValue());
		}
		
		for (Component child : this.children) {
			child.saveMaxConfiguration(configSourceName);
		}
		
		this.axis.moveTo(oldAxisValue);
	}
	
	public void setConfigurationVisibility(boolean visible, String configName, String configSourceName, 
				boolean overrideSourceVisibility, boolean recursive) {
		ConfigParams config = this.getConfigurationByName(configName);
		
		if (visible && !this.isRoot() && !this.parent.getConfigurationByName(configName).isVisible()) {
			parent.setConfigurationVisibility(visible, configName, configSourceName, overrideSourceVisibility, false);
		}
		
		config.setVisible(visible);
		try {
			ConfigParams configSource = this.getConfigurationByName(configSourceName);
			config.setRotation(configSource.getRotation());
			config.setTranslation(configSource.getTranslation());
			config.setScale(configSource.getScale());

			if (!overrideSourceVisibility) {
				config.setVisible(visible && configSource.isVisible());
			}
		} catch (NullPointerException e) {
			// Nothing to do : no source config
		} catch (Exception e) {
			Logger.getLogger("nomad-3d").severe(e.getMessage());
			e.printStackTrace();
		}
		
		if (recursive) {
			for (Component child : this.children) {
				child.setConfigurationVisibility(visible, configName, configSourceName, overrideSourceVisibility, recursive);
			}
		}
	}
	
	/**
	 * Gets a specific child by its name.
	 * @param name Name of the child
	 * @return Specified child if found, null otherwise
	 */
	public Component getChildByName(String name) {
		return getChildByName(name, false);
	}
	
	/**
	 * Gets a specific descendant by its name.
	 * @param name Name of the descendant
	 * @param recurse Recursively search in this compoennt's hierarchy or not
	 * @return Specified descendant if found, null otherwise
	 */
	public Component getChildByName(String name, boolean recurse) {
		for (Component child : this.children) {
			if (name.equals(child.getName())) {
				return child;
			}
			if (recurse) {
				Component grandChild = child.getChildByName(name, recurse);
				if (grandChild != null) {
					return grandChild;
				}
			}
		}
		return null;
	}
	
	/**
	 * Gets an ancestor by its name
	 * @param name name of the ancestor
	 * @return  Specified ancestor if found, null otherwise
	 */
	public Component getParentByName(String name) {
		if (this.parent == null) {
			return null;
		} else if (name.equals(this.parent.getName())) {
			return this.parent;
		} else {
			return this.parent.getParentByName(name);
		}
	}
	
	/**
	 * Gets a specific configuration.
	 * @param configName Name of the configuration
	 * @return The configuration if it exists, null otherwise
	 */
	public ConfigParams getConfigurationByName(String configName) {
		if (configName == null) {
			return null;
		}
		for (ConfigParams config : this.configurations) {
			if (configName.equals(config.getConfiguration())) {
				return config;
			}
		}
		return null;
	}
	
	/**
	 * Tests if this component belong to a component's hierarchy.
	 * @param comp Component
	 * @return true if this component is a descendant of comp
	 */
	public boolean isChild(Component comp) {
		if (this.parent == null) {
			return false;
		} else if (comp == this.parent) {
			return true;
		} else {
			return this.parent.isChild(comp);
		}
	}
	
	/**
	 * Tests if the component if a leaf.
	 * @return true if this component is a leaf
	 */
	public boolean isLeaf() {
		return this.children.isEmpty();
	}
	
	/**
	 * Tests if the component if the root.
	 * @return true if this component if the root
	 */
	public boolean isRoot() {
		return (this.parent == null);
	}
	
	/**
	 * Computes the tree depth of the component.
	 * @return The tree depth
	 */
	public int treeDepth() {
		if (this.parent == null) {
			return 0;
		} else {
			return this.parent.treeDepth() + 1;
		}
	}
	
	/**
	 * Changes the visibilty of this component's visual axis.
	 * @param onOff true to set the axis visible
	 */
	public void visibleAxis(boolean onOff) {
		if (this.axis != null)
			this.axis.getVisualGroup().setVisible(onOff);
	}
	
	/**
	 * Resets the axis transform of this component and its hierarchy. Used to reset the movement applied by the user.
	 * @param recurse true to reset the movement of the sub-hierarchy
	 */
	public void resetAxisTransform(boolean recurse) {
		this.axis.resetMovement();
		if (recurse) {
			for (Component child : this.children) {
				child.resetAxisTransform(recurse);
			}
		}
	}
	
	/**
	 * Sets the axis transform of this component and its hierarchy. Used to set the movement to zero.
	 * @param recurse true to reset the movement of the sub-hierarchy
	 */
	public void setToZeroAxisTransform(boolean recurse) {
		this.axis.setToZeroMovement();
		if (recurse) {
			for (Component child : this.children) {
				child.setToZeroAxisTransform(recurse);
			}
		}
	}
	
	/**
	 * Set the component's axis as fixed.
	 * @param recursive true to fix the hierarchy
	 */
	public void fix(boolean recursive) {
		this.setAxis(Axis.FIXED_AXIS.clone());
		if (recursive) {
			for (Component child : this.children) {
				child.fix(recursive);
			}
		}
	}

	/**
	 * Set the component's axis as free.
	 * @param recursive true to free the hierarchy
	 */
	public void free(boolean recursive) {
		this.setAxis(Axis.FREE_AXIS.clone());
		if (recursive) {
			for (Component child : this.children) {
				child.free(recursive);
			}
		}
	}
	
	/**
	 * Toggles the walls visibility.
	 * @param visible true to show the walls, false to hide them
	 * @param configName Name of the configuration to show
	 */
	public void setWallsVisible(boolean visible, String configName) {
		if (this.isWall()) {
			this.sceneNode.setVisible(visible && getConfigurationByName(configName).isVisible());
		}
		for (Component child : this.children) {
			child.setWallsVisible(visible, configName);
		}
	}
	
	/**
	 * Toggles the focus of the selection.
	 * @param selection The selection
	 * @param focusSelection true to see the selection only, false to see every component
	 * @param configName Name of the configuration to display
	 * @param recursive true to toggle the focus selection of the descendants
	 */
	public void focusSelection(CopyOnWriteArraySet<Component> selection, boolean focusSelection, String configName, boolean recursive) {
		this.setVisible(!this.isLeaf() || (focusSelection && selection.contains(this)) || !focusSelection, 
				configName, 
				false);
		if (recursive) {
			for (Component child : this.children) {
				child.focusSelection(selection, focusSelection, configName, recursive);
			}
		}
	}
	
	/**
	 * Sets the visibility of the component.
	 * @param visible New visibility
	 * @param configName Name of the configuration to display
	 * @param recursive true to apply the same visibility to the sub-hierarchy
	 */
	public void setVisible(boolean visible, String configName, boolean recursive) {
		this.sceneNode.setVisible(visible && getConfigurationByName(configName).isVisible());
		if (recursive) {
			for (Component child : this.children) {
				child.setVisible(visible, configName, recursive);
			}
		}
	}
	
	@Override 
	public String toString() {
		String str = "Component [\n";
		str += infoString();
		for (int i = 0 ; i < this.children.size() ; i++) {
			str += "Child " + i + " : " + this.children.get(i) + "\n";
		}
		str += "]";
		return str;
	}
	
	/**
	 * Compact string representation of the component
	 * @return Info string of the component
	 */
	public String infoString() {
		String str = "";
		str += "Name : " + this.name + "\n";
		str += "File name : " + this.fileName + "\n";
		str += "Configurations : \n";
		for (ConfigParams config : this.configurations) {
			str += config.toString() + "\n";
		}
		str += "Axis : " + this.axis + "\n";
		str += this.children.size() + " children\n";
		return str;
	}
	
	/**
	 * Size of the mesh(es) belonging to this component or its subhierarchy. 
	 * @param recursive true to count the triangles of the descendants
	 * @return Triangle count of the component's mesh(es)
	 */
	public long meshesSize(boolean recursive) {
		long size = 0;
		try {
			MeshView meshView = (MeshView)this.sceneNode;
			TriangleMesh mesh = (TriangleMesh)meshView.getMesh();
			size += mesh.getFaces().size();
		} catch (Exception e) {
			// Nothing to do
		}
		if (recursive) {
			for (Component child : this.children) {
				size += child.meshesSize(recursive);
			}
		}
		return size;
	}
	
	/**
	 * Hierarchy size, inluding this component.
	 * @return Number of component of the sub-hierarchy
	 */
	public long hierarchySize() {
		long size = 1;
		for (Component child : this.children) {
			size += child.hierarchySize();
		}
		return size;
	}
	
	/**
	 * Number of translational or rotational axes in the component's sub-hierarchy. 
	 * @return Axes count
	 */
	public long axesCount() {
		return axesCount(true);
	}

	
	/**
	 * Number of translational or rotational axes of the component's. 
	 * @param recursive true to count the axes of the descendants
	 * @return Axes count
	 */
	public long axesCount(boolean recursive) {
		long count = (this.axis.getType() == Axis.Type.TRANSLATION || this.axis.getType() == Axis.Type.ROTATION) ? 1 : 0;
		if (recursive) {
			for (Component child : this.children) {
				count += child.axesCount(recursive);
			}
		}
		return count;
	}
	
	/**
	 * Roots of a set of components, ie the components whose parent is not selected.
	 * @param components Set of components of a model
	 * @return Set of roots 
	 */
	public static HashSet<Component> rootsOf(Set<Component> components) {
		HashSet<Component> roots = new HashSet<>();
		for (Component c : components) {
			if (c.getParent() == null || !components.contains(c.getParent())) {
				roots.add(c);
			}
		}
		return roots;
	}
	
	public static Component ancestorOf(Set<Component> components) {
		if (components.isEmpty()) {
			return null;
		}
		Iterator<Component> iter = components.iterator();
		Component curComp = iter.next();
		if (curComp.isRoot()) {
			return null;
		}
		Component ancestor = curComp.getParent();
		while (iter.hasNext()) {
			curComp = iter.next();
			if (curComp.isRoot()) {
				return null;
			}
			while (!curComp.isChild(ancestor)) {
				if (ancestor.isRoot()) {
					return null;
				}
				ancestor = ancestor.getParent();
			}
		}
		return ancestor;
	}
	
	/**
	 * Computes the axes of this component and its sub-hierarchy for each configuration, merges the results and 
	 * analyzes the configurations in order to set the min/median/max values.
	 */
	void computeAxes() {
		if (!this.isRoot()) {
			Logger.getLogger("nomad-3d").warning(this.name + " is not the root");
		}
		for (ConfigParams config : this.configurations) {
			computeAxes(config.getConfiguration());
			storeAxes();
			resetAxes();
		}
		mergeConfigsAxes();
		configurationsAnalysis();
	}
	
	/**
	 * Compute the scene tree of this component's hierarchy, for the specified configuration.
	 * @param model Model owning the component
	 * @param configName Name of the configuration
	 * @param lod Level of detail of the displayed mesh
	 */
	void computeSceneHierarchy(Model model, String configName, int lod) {
		ConfigParams config = getConfigurationByName(configName);
		if (config == null) {
			Logger.getLogger("nomad-3d").severe("Cannot find configuration " + configName + " in component " + this.name);
			return;
		}
		
		Transform invParentTransform = null;
		Group parentGroup = null;
		
		if (this.parent != null) {
			parentGroup = (Group)this.parent.getSceneNode();
			try {
				invParentTransform = this.parent.getSceneNode().getLocalToSceneTransform().createInverse();
			} catch (Exception e) {
				e.printStackTrace();
			}
		} else {
			invParentTransform = new Affine();
		}

		if (this.children.isEmpty()) {
			String stlPath = this.fileName + ".STL";
			File stlFile = new File(model.getDirectoryPath() + File.separator + model.getGeometryDirectory(lod), stlPath);
			MeshView meshView = importMesh(model, stlFile);
			this.setSceneNode(meshView);
		} else {
			Group targetGroup = new Group();
			this.setSceneNode(targetGroup);
		}
		
		if (this.sceneNode != null) {
			if (parentGroup != null) {
				parentGroup.getChildren().add(this.sceneNode);
				this.axis.setSceneToParent(invParentTransform);
				parentGroup.getChildren().add(this.axis.getVisualGroup());
			} else {
				((Group)this.sceneNode).getChildren().add(this.axis.getVisualGroup());
			}
			
			this.sceneNode.setVisible(config.isVisible());
			this.sceneNode.getTransforms().addAll(invParentTransform, this.axis.getMovementTransform(), config.getTranslation(), config.getRotation(), config.getScale());
		}
		
		for (Component child : this.children) {
			child.computeSceneHierarchy(model, configName, lod);
		}
	}
	
	/**
	 * Imports the mesh of the component.
	 * @param model Model owning the component
	 * @param stlFile Path of the STL file containing the mesh
	 * @return The mesh view containing the imported mesh
	 */
	private MeshView importMesh(Model model, File stlFile) {
		TriangleMesh mesh = null;
		
		// Import
		try {
			if (stlFile.exists() && stlFile.canRead()) {
				StlMeshImporter importer = new StlMeshImporter();
				importer.read(stlFile);
				mesh = importer.getImport();
				mesh.setVertexFormat(VertexFormat.POINT_TEXCOORD);
				importer.close();
			} else if (!stlFile.exists()) {
				throw new IOException(this.name + " : the file does not exist");
			} else {
				throw new IOException(this.name + " : unable to read the file");
			}
		} catch (Exception e) {
			Logger.getLogger("nomad-3d").severe("Error loading file \"" + stlFile.getPath() + "\" in directory \"" 
					+ model.getDirectoryPath() + "\" : \n" + e.getMessage() + "\n");
			e.printStackTrace();
			return null;
		}

		MeshView meshView = new MeshView(mesh);
		meshView.setCullFace(CullFace.NONE);
		meshView.setMaterial(material);
		
		Component that = this;
		meshView.setOnMousePressed(new EventHandler<MouseEvent>() {
			@Override 
			public void handle(MouseEvent me) {
				if (!me.isPrimaryButtonDown()) return;
				
				mousePosX = me.getSceneX();
				mousePosY = me.getSceneY();
				mouseOldX = me.getSceneX();
				mouseOldY = me.getSceneY();

				if (me.isControlDown()) {
					if (model.getSelection().contains(that)) {
						model.removeFromSelection(that);
					} else {
						model.addToSelection(that);
					}
				}

				movingComponent = that;
				while (movingComponent != null 
					&& (movingComponent.getAxis().getType() == Axis.Type.FIXED 
						|| movingComponent.getAxis().getType() == Axis.Type.NONE)) {
					movingComponent = movingComponent.getParent();
				}
			}
		});
		meshView.setOnMouseClicked(new EventHandler<MouseEvent>() {
			@Override
			public void handle(MouseEvent me) {
				if (!me.isPrimaryButtonDown()) return;
				
				// Reset at the end of drag
				movingComponent = null;
			}
		});
		meshView.setOnMouseDragged(new EventHandler<MouseEvent>() {
			@Override
			public void handle(MouseEvent me) {
				if (!me.isPrimaryButtonDown()) return;
				if (movingComponent == null) return;
				
				mouseOldX = mousePosX;
				mouseOldY = mousePosY;
				mousePosX = me.getSceneX();
				mousePosY = me.getSceneY();
				Point2D delta = new Point2D((mousePosX - mouseOldX), (mousePosY - mouseOldY));
				
				double multiplier = 1;
//				if (me.isControlDown()) {
//					multiplier *= Nomad3DEditor.CONTROL_MULTIPLIER;
//				} 
//				if (me.isShiftDown()) {
//					multiplier *= Nomad3DEditor.SHIFT_MULTIPLIER;
//				}

				Point3D localAxisDir = movingComponent.getSceneNode().sceneToLocal(movingComponent.getAxis().getDirection());
				localAxisDir = localAxisDir.subtract(movingComponent.getSceneNode().sceneToLocal(Point3D.ZERO));
				Point2D screenAxisDir =  movingComponent.getSceneNode().localToScreen(localAxisDir);
				screenAxisDir =  screenAxisDir.subtract(movingComponent.getSceneNode().localToScreen(Point3D.ZERO));
				
				movingComponent.getAxis().move(delta, screenAxisDir, multiplier);
			}
		});
		meshView.setOnMouseMoved(new EventHandler<MouseEvent>() {
			@Override
			public void handle(MouseEvent me) {
				if (selectOn) {
					model.addToSelection(that);
				} else if (unselectOn) {
					model.removeFromSelection(that);
				}
			}
		});
		
		return meshView;
	}
	
	/**
	 * Compute the scene tree of this component's hierarchy, for the first configuration in its configuration list.
	 * @param model Model owning the component
	 * @param lod Level of detail
	 */
	void computeSceneHierarchy(Model model, int lod) {
		computeSceneHierarchy(model, this.configurations.get(0).getConfiguration(), lod);
	}
	
	/**
	 * Updates the visualization of the axes.
	 * @param selection Selected components
	 * @param enabled Activation of the axes visualization
	 */
	void updateVisualAxes(CopyOnWriteArraySet<Component> selection, boolean enabled) {
		for (Component child : this.children) {
			child.updateVisualAxes(selection, enabled);
		}
		this.axis.getVisualGroup().setVisible(enabled && selection.contains(this));
	}
	
	/**
	 * Set the component as selected
	 * @param selected true to select the component, false to not
	 * @param model Model owning the component
	 * @param recursive true to select the descendants, false to not
	 */
	void setSelected(boolean selected, Model model, boolean recursive) {
		if (selected) {
			model.addToSelection(this);
		} else {
			model.removeFromSelection(this);
		}
		if (recursive) {
			for (Component child : this.children) {
				child.setSelected(selected, model, recursive);
			}
		}
	}
	
	/**
	 * Shows a specific configuration
	 * @param configName name of the configuration
	 */
	void showConfiguration(String configName) {
		ConfigParams config = getConfigurationByName(configName);
		this.axis.setValue(config.getAxisValue());
		Transform invParentTransform = null;
		if (this.parent != null) {
			try {
				invParentTransform = this.parent.getSceneNode().getLocalToSceneTransform().createInverse();
			} catch (Exception e) {
				e.printStackTrace();
			}
		} else {
			invParentTransform = new Affine();
		}
		
		Node node = this.sceneNode;
		node.getTransforms().clear();
		node.getTransforms().addAll(
				invParentTransform, 
				this.axis.getMovementTransform(), 
				config.getTranslation(), 
				config.getRotation(), 
				config.getScale()
			);
		node.setVisible(config.isVisible());
		for (Component child : this.children) {
			child.showConfiguration(configName);
		}
	}
	
	/**
	 * Creates an item for the tree view of the editor's GUI.
	 * @param model Model containing this component.
	 * @return The created item
	 */
	CheckBoxTreeItem<Component> toTreeItem(Model model) {
		this.treeItem = new CheckBoxTreeItem<Component>(this);
		this.treeItem.setExpanded(this.isRoot());
		Component curComponent = this;
		this.treeItem.selectedProperty().addListener(new ChangeListener<Boolean>() {
			@Override
			public void changed(ObservableValue<? extends Boolean> observable, Boolean oldValue, Boolean newValue) {
				if (newValue.booleanValue()) {
					model.addToSelection(curComponent);
				} else {
					model.removeFromSelection(curComponent);
				}
			}
		});
		for (Component child : this.children) {
			this.treeItem.getChildren().add(child.toTreeItem(model));
		}
		return this.treeItem;
	}
	
	/**
	 * Analyzes the configuration to get the min/median/max values of the axis.
	 * @return true if the analysis is successful, false in case of errors
	 */
	boolean configurationsAnalysis() {		
		final int MEDIAN = 0, MIN = 1, MAX = 2, NB_CONFIGS = 3;
		ConfigParams[] configs = { null, null, null };
		Rotate[] aa = { null, null, null };
		String[] configNames = { "median", "min", "max" };
		
		// Check configurations
		boolean abort = false;
		for (int i = 0 ; i < NB_CONFIGS ; i++) {
			for (ConfigParams c : this.configurations) {
				if (c.getConfiguration().toLowerCase().equals(configNames[i])) {
					configs[i] = c;
					break;
				}
			}
			if (configs[i] == null) {
				Logger.getLogger("nomad-3d").warning(this.name + " : Cannot find the '" + configNames[i] + "' configuration, aborting configuration analysis \n");
				abort = true;
			}
		}
		if (abort) {
			return false;
		}
		
		// min < max
		if (configs[MIN].getAxisValue() > configs[MAX].getAxisValue()) {
			ConfigParams tmp = configs[MIN];
			configs[MIN] = configs[MAX];
			configs[MAX] = tmp;
		}

		// Check if min < median < max
		if (configs[MIN].getAxisValue() > configs[MEDIAN].getAxisValue() || configs[MAX].getAxisValue() < configs[MEDIAN].getAxisValue()) {
			Logger.getLogger("nomad-3d").warning("ERROR : the " + configNames[MEDIAN] + " configuration must be between " 
					+ configNames[MIN] + " and " + configNames[MAX] + ", aborting configuration analysis");
			return false;
		}
		
		// Compute boundaries
		switch (this.axis.getType()) {
			case ROTATION:
				for (int i = 0 ; i < NB_CONFIGS ; i++) {
					aa[i] = configs[i].axisAngleRotation();
					if (this.axis.getDirection().crossProduct(aa[i].getAxis()).magnitude() > EPSILON) {
						Logger.getLogger("nomad-3d").warning(this.name + " : rotation axis : axis of the configuration '" + configNames[i] + "' is not aligned with the rotation axis, aborting configuration analysis");
						return false;
					}
					if (this.axis.getDirection().dotProduct(aa[i].getAxis()) < 0) {
						// Invert the axis so everything is aligned and in the same direction
						aa[i].setAxis(aa[i].getAxis().multiply(-1));
						aa[i].setAngle(-aa[i].getAngle());
					}
				}
				this.axis.setMinValue(aa[MIN].getAngle());
				this.axis.setMaxValue(aa[MAX].getAngle());
				this.axis.setZeroValue(aa[MEDIAN].getAngle());
				break;
			case TRANSLATION:
				Point3D med = new Point3D(configs[MEDIAN].getTranslation().getTx(), configs[MEDIAN].getTranslation().getTy(), configs[MEDIAN].getTranslation().getTz());
				Point3D min = new Point3D(configs[MIN].getTranslation().getTx(), configs[MIN].getTranslation().getTy(), configs[MIN].getTranslation().getTz());
				Point3D max = new Point3D(configs[MAX].getTranslation().getTx(), configs[MAX].getTranslation().getTy(), configs[MAX].getTranslation().getTz());
				Point3D medMin = min.subtract(med);
				Point3D medMax = max.subtract(med);
				if (medMin.crossProduct(medMax).magnitude() > EPSILON) {
					Logger.getLogger("nomad-3d").warning(this.name + " : translation axis : positions of the configuration are not aligned, aborting configuration analysis");
					return false;
				}
				if (medMin.crossProduct(this.axis.getDirection()).magnitude() > EPSILON) {
					Logger.getLogger("nomad-3d").warning(this.name + " : translation axis : direction of the configuration is not the same as the direction of the axis, aborting configuration analysis");
					return false;
				}
				this.axis.setMinValue(min.dotProduct(this.axis.getDirection()));
				this.axis.setMaxValue(max.dotProduct(this.axis.getDirection()));
				this.axis.setZeroValue(med.dotProduct(this.axis.getDirection()));
				break;
			default:
				// Nothing to do
				break;
		}
		
		// Children
		for (Component child : this.children) {
			boolean success = child.configurationsAnalysis();
			if (!success) {
				return false;
			}
		}
		return true;
	}
	
	/**
	 * resets the movement along the axis for this component and its sub-hierarchy.
	 */
	private void resetAxes() {
		this.setAxis(this.initialAxis);
		for (Component child : this.children) {
			child.resetAxes();
		}
	}
	
	/**
	 * Store the axes of the hierarchy in the computed axes array of the components.
	 */
	private void storeAxes() {
		this.computedAxes.add(this.axis);
		for (Component child : this.children) {
			child.storeAxes();
		}
	}
	
	/**
	 * Merges the computed axes into the final axis of the component, for each component of the hierarchy.
	 */
	private void mergeConfigsAxes() {
		Axis finalAxis = null;
		boolean fixed = false;
		for (Axis axis : this.computedAxes) {
			if (axis.getType() == Axis.Type.TRANSLATION || axis.getType() == Axis.Type.ROTATION) {
				finalAxis = axis;
				break;
			} else if (axis.getType() == Axis.Type.FIXED) {
				fixed = true;
			}
		}
		if (finalAxis == null && fixed) {
			finalAxis = Axis.FIXED_AXIS.clone();
		} else if (finalAxis == null) {
			finalAxis = Axis.FREE_AXIS.clone();
		}
		this.setAxis(finalAxis);
		
		for (Component child : this.children) {
			child.mergeConfigsAxes();
		}
	}

	/**
	 * Computes the axes of this component and its hierarchy.
	 * @param configName Name of the configuration
	 */
	private void computeAxes(String configName) {
		ConfigParams config = getConfigurationByName(configName);
		if (config == null) {
			Logger.getLogger("nomad-3d").severe("Cannot find configuration " + configName + " in component " + this.name);
			return;
		}
		
		// Compute all axes in this subtree
		boolean needComputation = false;
		for (Component child : this.children) {
			child.computeAxes(configName);
			needComputation = needComputation || (child.getAxis().getMethod() == Axis.Method.TO_COMPUTE);
		}
		if (!needComputation) {
			return;
		} else if (config.getMateGroup().isEmpty()) {
			for (Component child : this.children) {
				if (child.getAxis().getMethod() == Axis.Method.TO_COMPUTE) {
					child.setAxis(Axis.FREE_AXIS.clone());
				}
			}
			return;
		}

		// Block pass
		blockPass(config);
		
		// Axis pass
		axisPass(config);
		
		// Clean hierarchy when all the children were in the same block (different than root block)
		if (this.children.size() == 1 && this.children.get(0).getFileName().equals("NOMAD_3D_INSERTED")
				&& (this.axis.getType() == Axis.Type.NONE || this.axis.getType() == Axis.Type.FIXED)) {
			Component onlyInsertedChild = this.children.get(0);
			this.setAxis(onlyInsertedChild.getAxis());
			this.removeChild(onlyInsertedChild);
			for (Component child : onlyInsertedChild.getChildren()) {
				this.children.add(child);
				child.parent = this;
			}
			onlyInsertedChild.getChildren().clear();
		}
	}
	
	/**
	 * Block pass of the axes computation.
	 * @param config Current configuration
	 */
	private void blockPass(ConfigParams config) {
		HashSet<Component> rootBlock = computeRootBlock(config);
		LinkedList<HashSet<Component>> blocks = computeBlocks(rootBlock, config);
		regroupBlocks(blocks);
		
		// Root block marked as fixed
		for (Component comp : rootBlock) {
			if (comp == this) {
				continue;
			}
			Axis fixedAxis = new Axis();
			fixedAxis.setType(Axis.Type.FIXED);
			fixedAxis.setMethod(Axis.Method.COMPUTED);
			comp.setAxis(fixedAxis);
		}
	}

	/**
	 * Axis pass of the axes computation.
	 * @param config Current configuration
	 */
	private void axisPass(ConfigParams config) {
		Stack<Component> computedComponents = new Stack<Component>();
		for (Component child : this.children) {
			if (child.getAxis().getMethod() != Axis.Method.TO_COMPUTE) {
				computedComponents.push(child);
			}
		}
		computedComponents.push(this);
		
		while (!computedComponents.isEmpty()) {
			Component[] childs = { computedComponents.pop(), null };
			Axis[] dofs = { childs[0].getAxis(), null };
			
			for (int j = 0 ; j < this.children.size() ; j++) {
				childs[1] = this.children.get(j);
				if (childs[0] == childs[1] || childs[1].getAxis().getMethod() != Axis.Method.TO_COMPUTE) {
					continue;
				}

				// Mates between children (blocks are regrouped as new children so it takes them into account)
				HashSet<Mate> curMates = config.matesBetween(childs[0], childs[1]);

				// Axis of 1, relative to 0
				dofs[1] = Axis.fromMates(curMates, config);

				Axis computedAxis = axisFromDofs(childs, dofs);
				if (computedAxis != null) {
					childs[1].setAxis(computedAxis);
					computedComponents.push(childs[1]);
				}
			}
		}
		
		// Unreached children marked as free components
		for (Component child : this.children) {
			if (child.getAxis().getMethod() == Axis.Method.TO_COMPUTE) {
				child.setAxis(Axis.FREE_AXIS.clone());
			}
		}
	}
	
	/**
	 * Computes the root block during the block pass.
	 * @param config Current configuration
	 * @return The root block
	 */
	private HashSet<Component> computeRootBlock(ConfigParams config) {
		HashSet<Component> rootBlock = new HashSet<Component>();
		rootBlock.add(this);
		for (Component child : this.children) {
			if (child.getAxis().getMethod() != Axis.Method.TO_COMPUTE && child.getAxis().getType() == Axis.Type.FIXED) {
				rootBlock.add(child);
				continue;
			}
			HashSet<Mate> curMates = config.matesBetween(this, child);
			Axis curDof = Axis.fromMates(curMates, config);
			
			if (curDof.getType() == Axis.Type.FIXED) {
				rootBlock.add(child);
			}
		}
		return rootBlock;
	}
	
	/**
	 * Computes the blocks, apart from the root block, during the block pass.
	 * @param rootBlock Root block
	 * @param config Current configuration
	 * @return List of the blocks
	 */
	private LinkedList<HashSet<Component>> computeBlocks(HashSet<Component> rootBlock, ConfigParams config) {
		LinkedList<HashSet<Component>> blocks = new LinkedList<HashSet<Component>>();
		Component[] childs = { null, null };
		for (int i = 0 ; i < this.children.size() ; i++) {
			childs[0] = this.children.get(i);
			for (int j = 0 ; j < this.children.size() ; j++) {
				childs[1] = this.children.get(j);
				
				// Blocks of the current children (null if a child do not belong to any block)
				ArrayList<HashSet<Component>> curBlocks = new ArrayList<HashSet<Component>>();
				for (int k = 0 ; k < 2 ; k++) {
					curBlocks.add(null);
					if (rootBlock.contains(childs[k])) {
						curBlocks.set(k, rootBlock);
					} else {
						for (HashSet<Component> b : blocks) {
							if (b.contains(childs[k])) {
								curBlocks.set(k, b);
								break;
							}
						}
					}
				}
				
				// Mates between components or blocks 
				HashSet<Mate> curMates = new HashSet<>();
				if (curBlocks.get(0) != null && curBlocks.get(1) != null) {
					for (Component compBlock0 : curBlocks.get(0)) {
						for (Component compBlock1 : curBlocks.get(1)) {
							curMates.addAll(config.matesBetween(compBlock0, compBlock1));
						}
					}
				} else if (curBlocks.get(1) != null) {
					for (Component compBlock : curBlocks.get(1)) {
						curMates.addAll(config.matesBetween(childs[0], compBlock));
					}
				} else if (curBlocks.get(0) != null) {
					for (Component compBlock : curBlocks.get(0)) {
						curMates.addAll(config.matesBetween(compBlock, childs[1]));
					}
				} else {
					curMates.addAll(config.matesBetween(childs[0], childs[1]));
				}
				
				// Axis of 1, relative to 0
				Axis curDof = Axis.fromMates(curMates, config);
				
				// Blocks update : add, merge or create
				if (curDof.getType() == Axis.Type.FIXED) {					
					if (curBlocks.get(0) != null && curBlocks.get(1) != null) {
						// Merge blocks
						if (curBlocks.get(0) != curBlocks.get(1)) {
							if (curBlocks.get(1) == rootBlock) {
								curBlocks.get(1).addAll(curBlocks.get(0));
								blocks.remove(curBlocks.get(0));
							} else {
								curBlocks.get(0).addAll(curBlocks.get(1));
								blocks.remove(curBlocks.get(1));
							}
						}
					} else if (curBlocks.get(0) == null && curBlocks.get(1) != null) {
						// Add 0 to block of 1
						curBlocks.get(1).add(childs[0]);
					} else if (curBlocks.get(1) == null && curBlocks.get(0) != null) {
						// Add 1 to block of 0
						curBlocks.get(0).add(childs[1]);
					} else {
						// New block containing 0 and 1
						HashSet<Component> newBlock = new HashSet<Component>();
						newBlock.add(childs[0]);
						newBlock.add(childs[1]);
						blocks.add(newBlock);
					}
				}
			}
		}
		return blocks;
	}
	
	/**
	 * Regroup the blocks in new components. These are inserted as children of this components.
	 * @param blocks Computed blocks
	 */
	private void regroupBlocks(LinkedList<HashSet<Component>> blocks) {
		for (HashSet<Component> block : blocks) {
			if (block.size() < 2) {
				Logger.getLogger("nomad-3d").severe("[" + this.name + "] REGROUP BLOCKS : block of size " + block.size());
				continue;
			}
			
			insertBlock(block, true);
		}
	}
	
	/**
	 * Inserts a block as a child of this component.
	 * @param block Block to insert
	 * @param fixedBlock true to set the block as fixed
	 * @return The inserted component
	 */
	public Component insertBlock(HashSet<Component> block, boolean fixedBlock) {
		return insertBlock(block, fixedBlock, "INSERTED");
	}
	public Component insertBlock(HashSet<Component> block, boolean fixedBlock, String tag) {
		Component inserted = new Component();
		
		for (Iterator<Component> iter = block.iterator() ; iter.hasNext() ; ) {
			Component comp = iter.next();
			if (fixedBlock) {
				comp.setAxis(Axis.FIXED_AXIS.clone());
			}
			inserted.addChild(comp);
		}
		
		for (ConfigParams config : this.configurations) {
			ConfigParams newConfig = new ConfigParams(inserted);
			newConfig.setConfiguration(config.getConfiguration());
			inserted.addConfiguration(newConfig);
		}
		
		inserted.setName("NOMAD_3D_" + tag + "-" + insertedIndex);
		inserted.setFileName("NOMAD_3D_INSERTED");
		inserted.setAxis(Axis.FREE_AXIS.clone());
		inserted.getAxis().setMethod(Axis.Method.TO_COMPUTE);
		
		this.addChild(inserted);
		Component.insertedIndex++;
		
		return inserted;
	}
	
	/**
	 * Computes an absolute axis from relative ones.
	 * @param childs Pair of the concerned children
	 * @param dofs pair of axes. The first is absolute and the second is relative.
	 * @return Absolute axis of the second child, or null if it cannot be computed given these parameters
	 */
	private Axis axisFromDofs(Component[] childs, Axis[] dofs) {
		Axis computedAxis = null;
		
		if (dofs[1].getType() == Axis.Type.FIXED) { 
			Logger.getLogger("nomad-3d").severe("Fixed child undetected during the block pass - " + childs[1].getName() + " fixed to " + childs[0].getName());
		} else if ((dofs[0].getType() == Axis.Type.FIXED || childs[0] == this) && dofs[1].getType() != Axis.Type.NONE) {
			computedAxis = dofs[1].clone();
		} else if (dofs[0].getType() == Axis.Type.TRANSLATION) {
			switch (dofs[1].getType()) {
				case ROTATION:
					if (dofs[0].getDirection().crossProduct(dofs[1].getDirection()).magnitude() < EPSILON) {
 						computedAxis = dofs[1].clone(); // priority to rotation
//						computedAxis = dofs[0]; // priority to translation
					} else {
						childs[1].setAxis(dofs[1].clone());
						childs[0].addChild(childs[1]);
					}
					break;
				case TRANSLATION:
						childs[1].setAxis(dofs[1].clone());
						childs[0].addChild(childs[1]);
					break;
				default:
					break;
			}
		} else if (dofs[0].getType() == Axis.Type.ROTATION) {
			switch (dofs[1].getType()) {
				case ROTATION:
						childs[1].setAxis(dofs[1].clone());
						childs[0].addChild(childs[1]);
				case TRANSLATION:
					if (dofs[0].getDirection().crossProduct(dofs[1].getDirection()).magnitude() < EPSILON) {
						computedAxis = dofs[0].clone(); // priority to rotation
//						computedAxis = dofs[1]; // priority to translation
					} else {
						childs[1].setAxis(dofs[1].clone());
						childs[0].addChild(childs[1]);
					}
					break;
				default:
					break;
			}
		}
		
		return computedAxis;
	}
}
