package fr.ill.ics.n3d.model;

import java.util.ArrayList;
import java.util.HashSet;

import javafx.geometry.Point3D;
import javafx.scene.transform.Affine;
import javafx.scene.transform.Rotate;
import javafx.scene.transform.Scale;
import javafx.scene.transform.Transform;

/**
 * Class gathering the configuration-specific parameters of a component.
 * @author dages
 */
public class ConfigParams {
	/** Component owning this configuration. */
	private Component owner;
	
	/** Name of this configuration. */
	private String configuration;
	
	/** true is the component can move in this configuration, false if it cannot. */
	private boolean fixed;
	
	/** Visibility of the component in this configuration. */
	private boolean visible;
	
	/** Translation of the configuration, in global coordinates. */
	private Affine translation;
	
	/** Rotation of the configuration, in global coordinates. */
	private Affine rotation;
	
	/** Scale of the configuration, in global coordinates. */
	private Scale scale;
	
	/** Axis value of the configuration. */
	private double axisValue;
	
	/** Mate group of the configuration. */
	private ArrayList<Mate> mateGroup;

	/**
	 * Constructor.
	 * @param owner Owner of the configuration
	 */
	public ConfigParams(Component owner) {
		this.owner = owner;
		this.configuration = "";
		this.fixed = false;
		this.visible = true;
		this.translation = new Affine();
		this.rotation = new Affine();
		this.scale = new Scale();
		this.axisValue = 0.0;
		this.mateGroup = new ArrayList<>();
	}
	
	/**
	 * Gets the owner.
	 * @return The owner
	 */
	public Component getOwner() {
		return owner;
	}
	
	/**
	 * Gets the name of the configuration.
	 * @return The name
	 */
	public String getConfiguration() {
		return configuration;
	}
	
	/**
	 * gets the mobility of the configuration
	 * @return true if the component is fixed, false otherwise
	 */
	public boolean isFixed() {
		return fixed;
	}

	/**
	 * Gets the visibility of the configuration
	 * @return true if the component is visible, false otherwise
	 */
	public boolean isVisible() {
		return visible;
	}
	
	/**
	 * Gets the translation transform
	 * @return The translation transform
	 */
	public Affine getTranslation() {
		return translation;
	}
	
	/**
	 * Gets the rotation transform
	 * @return The rotation transform
	 */
	public Affine getRotation() {
		return rotation;
	}
	
	/**
	 * Gets the scale transform.
	 * @return The scale transform
	 */
	public Scale getScale() {
		return scale;
	}
	
	/**
	 * Gets the axis value.
	 * @return The axis value
	 */
	public double getAxisValue() {
		return axisValue;
	}
	
	/**
	 * Gets the mate group.
	 * @return The mate group
	 */
	public ArrayList<Mate> getMateGroup() {
		return mateGroup;
	}
	
	/**
	 * Sets the name of the configuration.
	 * @param configuration New name
	 */
	public void setConfiguration(String configuration) {
		this.configuration = configuration;
	}
	
	/**
	 * Sets the mobility of the configuration.
	 * @param fixed New mobility
	 */
	public void setFixed(boolean fixed) {
		this.fixed = fixed;
	}
	
	/**
	 * Sets the visibility of the configuration.
	 * @param visible New visibility
	 */
	public void setVisible(boolean visible) {
		this.visible = visible;
	}
	
	/**
	 * Sets the translation transform.
	 * @param translation New translation transform
	 */
	public void setTranslation(Affine translation) {
		this.translation = translation;
	}
	
	/**
	 * Sets the rotation transform.
	 * @param rotation New rotation transform
	 */
	public void setRotation(Affine rotation) {
		this.rotation = rotation;
	}
	
	/**
	 * Sets the scale transform.
	 * @param scale New scale transform
	 */
	public void setScale(Scale scale) {
		this.scale = scale;
	}
	
	/**
	 * Sets the translation and rotation transform, based on one affine transform
	 * @param transform Input transform
	 */
	public void setTransform(Transform transform) {
		Point3D translation = transform.transform(Point3D.ZERO);
		Point3D xRotation = transform.transform(Rotate.X_AXIS).subtract(translation);
		Point3D yRotation = transform.transform(Rotate.Y_AXIS).subtract(translation);
		Point3D zRotation = transform.transform(Rotate.Z_AXIS).subtract(translation);
		
		this.translation = new Affine();
		this.translation.setTx(translation.getX());
		this.translation.setTy(translation.getY());
		this.translation.setTz(translation.getZ());
		
		this.rotation = new Affine();
		this.rotation.setMxx(xRotation.getX());
		this.rotation.setMyx(xRotation.getY());
		this.rotation.setMzx(xRotation.getZ());
		this.rotation.setMxy(yRotation.getX());
		this.rotation.setMyy(yRotation.getY());
		this.rotation.setMzy(yRotation.getZ());
		this.rotation.setMxz(zRotation.getX());
		this.rotation.setMyz(zRotation.getY());
		this.rotation.setMzz(zRotation.getZ());
	}
	
	/**
	 * Sets the axis value.
	 * @param axisValue New axis value
	 */
	public void setAxisValue(double axisValue) {
		this.axisValue = axisValue;
	}
	
	/**
	 * Adds a mate to the configuration's mate group.
	 * @param mate New mate
	 */
	public void addMate(Mate mate) {
		this.mateGroup.add(mate);
	}
	
	/**
	 * Converts the rotation transform to an axis-angle perspective.
	 * @return The axis-angle rotation transform
	 */
	public Rotate axisAngleRotation() {
		// https://en.wikipedia.org/wiki/Rotation_formalisms_in_three_dimensions#Rotation_matrix_.E2.86.94_Euler_axis.2Fangle
		double epsilon = 1e-10;
		double theta = Math.acos(0.5 * (this.rotation.getMxx() + this.rotation.getMyy() + this.rotation.getMzz() - 1.0));
		double e1 = 0.0, e2 = 1.0, e3 = 0.0;
		if (Math.abs(theta) > epsilon) {
			double twoSin = 2.0 * Math.sin(theta);
			e1 = (this.rotation.getMzy() - this.rotation.getMyz()) / twoSin;
			e2 = (this.rotation.getMxz() - this.rotation.getMzx()) / twoSin;
			e3 = (this.rotation.getMyx() - this.rotation.getMxy()) / twoSin;
		}
		
		double angle = Math.toDegrees(theta);
		Point3D axis = new Point3D(e1, e2, e3).normalize();
		return new Rotate(angle, axis);
	}
	
	/**
	 * Gets all the mates between two components (hierarchies) in the mate group of this component.
	 * @param c1 First component
	 * @param c2 Second component
	 * @return Set of mates constraining the two components
	 */
	public HashSet<Mate> matesBetween(Component c1, Component c2) {
		HashSet<Mate> mates = new HashSet<Mate>();
		for (Mate mate : this.mateGroup) {
			Component[] targets = { mate.getEntity(0).getTarget(), mate.getEntity(1).getTarget() };
			boolean[] addMate = {false, false};
			boolean[] addHierarchy = {(c1 != this.owner), (c2 != this.owner)};
			addMate[0] = ((targets[0] == c1 || (addHierarchy[0] && targets[0].isChild(c1))) && (targets[1] == c2 || (addHierarchy[1] && targets[1].isChild(c2))));
			addMate[1] = ((targets[1] == c1 || (addHierarchy[0] && targets[1].isChild(c1))) && (targets[0] == c2 || (addHierarchy[1] && targets[0].isChild(c2))));
			if (addMate[0] || addMate[1]) {
				mates.add(mate);
			}
		}
		return mates;
	}

	@Override
	public String toString() {
		String str = "ConfigParams [\n";
		str += "\t Configuration : " + configuration + "\n";
		str += "\t Fixed=" + fixed + ", visible=" + visible + "\n";
		str += "\t Axis value : " + axisValue + "\n";
		str += "\t Mate group : " + mateGroup.size() + " mates \n";
		str += "]";
		return str;
	}
	
	
}
