package fr.ill.ics.n3d.model;

import java.util.logging.Logger;

import javafx.scene.transform.Transform;

/**
 * Class representing a SolidWorks mate exported with the Nomad 3D addin.
 * These mates are used to define cinematic constraints.
 * 
 * @see 
 * <a href="http://help.solidworks.com/2015/English/SolidWorks/sldworks/c_Mates_Overview_Assemblies.htm?id=3d28cd3d7b384e6ea3e565c5db2ca24f#Pg0&ProductType=&ProductName=">
 * SolidWorks documentation
 * </a>
 * 
 * @author dages
 */
public abstract class Mate {
	
	/**
	 * Alignment of a mate.
	 * @author dages
	 */
	public enum Alignment {
		/** Aligned mate. */
		ALIGNED("Aligned"),
		/** Anti-aligned mate, ie the mate is reversed. */
		ANTI_ALIGNED("AntiAligned"),
		/** Closest alignment. */
		CLOSEST("Closest");
		
		/** String representation of the alignment. */
		private String name;
		
		/**
		 * Constructor.
		 * @param name String representation of the alignment
		 */
		Alignment(String name) {
			this.name = name;
		}
		
		/**
		 * Converts a string to an alignment.
		 * @param name Name of the alignment
		 * @return The designated alignment if it exists, null otherwise
		 */
		public static Alignment fromString(String name) {
			switch (name) {
				case "Aligned":
					return Alignment.ALIGNED;
				case "AntiAligned":
					return Alignment.ANTI_ALIGNED;
				case "Closest":
					return Alignment.CLOSEST;
				default:
					Logger.getLogger("nomad-3d").severe("Unknown alignment \"" + name + "\"\n");
					return null;
			}
		}
		
		/**
		 * String representation of the alignment.
		 * @return Name of the alignment
		 */
		@Override
		public String toString() {
			return this.name;
		}
	}
	
	/** Threshold used for axis computation. */
	protected static final double EPSILON = 1e-1;
	
	/** Name of the mate. */
	private String name;
	
	/** Alignment of the mate. */
	private Alignment alignment;
	
	/** Minimum value of the mate. */
	private double min;
	
	/** Maximum value of the mate. */
	private double max;
	
	/** Whether the mate is flipped or not. */
	private boolean flipped;
	
	/** Entities of the mate. */
	private MateEntity[] entities;
	
	/** Number of mate entities for a mate. */
	public static final int NB_ENTITIES = 2;
	
	/**
	 * Default constructor.
	 */
	protected Mate() {
		this.name = "";
		this.alignment = Alignment.ALIGNED;
		this.min = 0;
		this.max = 0;
		this.flipped = false;
		this.entities = new MateEntity[NB_ENTITIES];
		for (int i = 0 ; i < NB_ENTITIES ; i++) {
			this.entities[i] = null;
		}
	}
	
	/**
	 * Gets the name.
	 * @return The name
	 */
	public String getName() {
		return this.name;
	}
	
	/**
	 * Gets the alignment.
	 * @return The alignment
	 */
	public Alignment getAlignment() {
		return alignment;
	}
	
	/**
	 * Gets the minimum value.
	 * @return The minimum value
	 */
	public double getMin() {
		return this.min;
	}

	/**
	 * Gets the maximum value.
	 * @return The maximum value
	 */
	public double getMax() {
		return this.max;
	}
	
	/**
	 * Gets the flipped status.
	 * @return true if the mate is flipped, false otherwise
	 */
	public boolean isFlipped() {
		return this.flipped;
	}
	
	/**
	 * Gets an entity of this mate.
	 * @param i Entity index, must be between 0 and NB_ENTITIES-1
	 * @return The entity
	 */
	public MateEntity getEntity(int i) {
		return this.entities[i];
	}
	
	/**
	 * Sets the name.
	 * @param name New name
	 */
	public void setName(String name) {
		this.name = name;
	}
	
	/**
	 * Sets the alignment.
	 * @param alignment New alignment
	 */
	public void setAlignment(Alignment alignment) {
		this.alignment = alignment;
	}
	
	/**
	 * Sets the minimum value.
	 * @param min New minimum
	 */
	public void setMin(double min) {
		this.min = min;
	}
	
	/**
	 * Sets the maximum value.
	 * @param max New maximum
	 */
	public void setMax(double max) {
		this.max = max;
	}
	
	/**
	 * Set the flipped tag.
	 * @param flipped New flipped
	 */
	public void setFlipped(boolean flipped) {
		this.flipped = flipped;
	}
	
	/**
	 * Sets an entity.
	 * @param i Entity index, must be between 0 and NB_ENTITIES-1
	 * @param entity New entity
	 */
	public void setEntity(int i, MateEntity entity) {
		this.entities[i] = entity;
	}
	
	/**
	 * Applies the mate constraint to translational and rotational degrees of freedom.
	 * @param translation Translation axis, represents translational DOFs
	 * @param rotation Rotation axis, represents rotational DOFs
	 * @param rootToScene Root to scene transform
	 */
	public abstract void apply(Axis translation, Axis rotation, Transform rootToScene);
	
	@Override
	public boolean equals(Object o) {
		boolean isEqual = false;
		if (o == this) {
			return true;
		} else if (o instanceof Mate) {
			Mate mate = (Mate)o;
			isEqual = true;
			isEqual &= this.name.equals(mate.name);
//			isEqual &= (this.type == mate.type);
			isEqual &= (this.min == mate.min);
			isEqual &= (this.max == mate.max);
			isEqual &= !(this.flipped ^ mate.flipped);
			if (this.entities[0] != null && this.entities[1] != null) {
				isEqual &= (this.entities[0].equals(mate.entities[0]) && this.entities[1].equals(mate.entities[1])) 
							|| (this.entities[0].equals(mate.entities[1]) && this.entities[1].equals(mate.entities[0]));
			}
		}
		return isEqual;
	}
	
	@Override
	public int hashCode() {
		int code = this.name.length();
		for (int i = 0 ; i < NB_ENTITIES ; i++) {
			if (this.entities[i] != null) {
				code += this.entities[i].hashCode();
			}
		}
		return code;
	}
	
	@Override
	public String toString() {
		String str = "Mate [ \n";
		str += "Name : " + this.name + "\n";
//		str += "Alignment : " + this.alignment + "\n";
//		str += "Min : " + this.min + "\n";
//		str += "Max : " + this.max + "\n";
//		str += "Flipped : " + this.flipped + "\n";
		for (int i = 0 ; i < NB_ENTITIES ; i++) {
			str += "Entity " + i + " : " + this.entities[i] + "\n";
		}
		str += "]";
		return str;
	}
}
