package fr.ill.ics.n3d.model;

import javafx.geometry.Point3D;
import javafx.scene.transform.Transform;

/**
 * Class representing an entity constrained by a mate.
 * @see <a
 *      href="http://help.solidworks.com/2015/english/api/sldworksapi/SolidWorks.Interop.sldworks~SolidWorks.Interop.sldworks.IMateEntity2~EntityParams.html">
 *      SolidWorks documentation </a>
 * @author dages
 */
public class MateEntity {
	
	/**
	 * Type of an entity.
	 * These types are a subset of SolidWorks mate entities types.
	 * @author dages
	 */
	public enum Type {
		/** Point constraint. */
		POINT("Point"),
		/** Line constraint. */
		LINE("Line"),
		/** Plane constraint. */
		PLANE("Plane"),
		/** Cylinder constraint. */
		CYLINDER("Cylinder"),
		/** Circle constraint. */
		CIRCLE("Circle"),
		/** Special value for unsupported constraints. */
		UNSUPPORTED("Unsupported"),
		/** Special value for unknown constraints. */
		UNKNOWN("UNKNOWN");
		
		/** String representation of the type. */
		private String name;
		
		/**
		 * Constructor.
		 * @param name String representation of the type
		 */
		Type(String name) {
			this.name= name;
		}
		
		/**
		 * Converts a string to a type.
		 * @param name String representation of the type
		 * @return The type if it exists, null otherwise
		 */
		public static Type fromString(String name) {
			switch (name) {
				case "Point":
					return Type.POINT;
				case "Line":
					return Type.LINE;
				case "Plane":
					return Type.PLANE;
				case "Cylinder":
					return Type.CYLINDER;
				case "Circle":
					return Type.CIRCLE;
				case "Unsupported":
					return Type.UNSUPPORTED;
				default:
					System.err.println("MateEntity.Type.fromString : ERROR : unknown type \"" + name + "\"");
					return Type.UNKNOWN;
			}
		}
		
		/**
		 * String representation of the type.
		 * @return Name of the type
		 */
		@Override
		public String toString() {
			return this.name;
		}
	}
	
	/** Type of the entity. */
	private Type type;
	
	/** Component constrained by the entity. */
	private Component target;
	
	/** Position of the constraint, in local system. */
	private Point3D position;
	
	/** Axis (or direction) of the constraint, in local system. Should be a unit vector. */
	private Point3D axis;
	
	/** Array of radius of the constraint. */
	private double[] radius;
	
	/** Number of radius per entity. */
	public static final int NB_RADIUS = 2;

	/**
	 * Default constructor.
	 */
	public MateEntity() {
		this.type = Type.UNKNOWN;
		this.target = null;
		this.position = new Point3D(0, 0, 0);
		this.axis = new Point3D(0, 0, 0);
		this.radius = new double[NB_RADIUS];
		for (int i = 0 ; i < NB_RADIUS ; i++) {
			this.radius[i] = 0;
		}
	}
	
	/**
	 * Gets the type.
	 * @return The type
	 */
	public Type getType() {
		return this.type;
	}
	
	/**
	 * Gets the target.
	 * @return The target.
	 */
	public Component getTarget() {
		return this.target;
	}
	
	/**
	 * Gets the position.
	 * @return The position.
	 */
	public Point3D getPosition() {
		return this.position;
	}
	
	/**
	 * Gets the axis.
	 * @return The axis.
	 */
	public Point3D getAxis() {
		return this.axis;
	}
	
	/**
	 * Gets the radius.
	 * @param i Radius index
	 * @return The radius.
	 */
	public double getRadius(int i) {
		return this.radius[i];
	}
	
	/**
	 * Sets the type.
	 * @param type The type
	 */
	public void setType(Type type) {
		this.type = type;
	}
	
	/**
	 * Sets the target.
	 * @param target The target
	 */
	public void setTarget(Component target) {
		this.target = target;
	}
	
	/**
	 * Sets the position.
	 * @param position The position
	 */
	public void setPosition(Point3D position) {
		this.position = position;
	}
	
	/**
	 * Sets the axis.
	 * @param axis The axis
	 */
	public void setAxis(Point3D axis) {
		this.axis = axis;
	}
	
	/**
	 * Sets the radius.
	 * @param i Radius index
	 * @param radius The radius
	 */
	public void setRadius(int i, double radius) {
		this.radius[i] = radius;
	}
	
	/**
	 * Position of the constraint in global system.
	 * @return The global position
	 */
	public Point3D positionInScene() {
		Transform toScene = this.target.getSceneNode().getLocalToSceneTransform();
		return toScene.transform(this.position);
	}
	
	/**
	 * Axis (or direction) of the constraint in global system.
	 * @return The global axis
	 */
	public Point3D axisInScene() {
		Transform toScene = this.target.getSceneNode().getLocalToSceneTransform();
		Point3D origin = toScene.transform(Point3D.ZERO);
		Point3D transformedAxis = toScene.transform(this.axis);
		return transformedAxis.subtract(origin);
	}

	@Override
	public boolean equals(Object o) {
		boolean isEqual = false;
		if (o == this) {
			return true;
		} else if (o instanceof MateEntity) {
			MateEntity entity = (MateEntity)o;
			isEqual = true;
			isEqual &= (this.type == entity.type);
			isEqual &= (this.target == entity.target);
			isEqual &= this.position.equals(entity.position);
			isEqual &= this.axis.equals(entity.axis);
			isEqual &= (this.radius[0] == entity.radius[0]);
			isEqual &= (this.radius[1] == entity.radius[1]);
		}
		return isEqual;
	}
	
	@Override
	public int hashCode() {
		int code = this.type.toString().length();
		if (target != null) {
			code += this.target.getName().length();
		}
		return code;
	}
	
	/**
	 * Returns a string representation of the object.
	 * @return String representation of the object.
	 */
	@Override
	public String toString() {
		String str = "MateEntity [\n ";
		str += "Type : " + this.type + "\n ";
		str += "Target : " + this.target.getName() + "\n ";
		str += "Position : " + this.position + "\n";
		str += "Axis : " + this.axis + "\n";
		str += "Radius : ";
		for (int i = 0 ; i < NB_RADIUS ; i++) {
			str += radius[i] + " ";
		}
		str += "\n";
		str += "]";
		return str;
	}
}
