package fr.ill.ics.n3d.model;

import java.util.HashMap;
import java.util.concurrent.CopyOnWriteArraySet;
import java.util.logging.Logger;

import javafx.scene.paint.Color;
import javafx.scene.paint.PhongMaterial;
import javafx.scene.shape.MeshView;


/**
 * Class describing a Nomad 3D model.
 * @author dages
 */
public class Model {
	
	/**
	 * States of a model during the preprocess pipeline
	 * @author dages
	 */
	public enum State {
		/** Exported model from SolidWorks, needs to be converted. */
		EXPORTED("exported") {
			@Override
			public State next() {
				return CONVERTED;
			}
		},
		/** Converted model. Can be edited. */
		CONVERTED("converted") {
			@Override
			public State next() {
				return EDITED;
			}
		},
		/** Edited model. */
		EDITED("edited") {
			@Override
			public State next() {
				return EDITED;
			}
		};
		
		/** name of the state. */
		private String name;
		
		/**
		 * Constructor
		 * @param name Name of the state
		 */
		State(String name) {
			this.name = name;
		}
		
		/** 
		 * Next state in the pipeline.
		 * @return The next state
		 */
		public abstract State next();
		
		/**
		 * Converts a string to a state
		 * @param name Name of the state
		 * @return The state if it exists, null otherwise
		 */
		public static State fromString(String name) {
			switch (name) {
				case "exported":
					return State.EXPORTED;
				case "converted":
					return State.CONVERTED;
				case "edited":
					return State.EDITED;
				default:
					return null;
			}
		}
		
		/**
		 * String representation of the state.
		 * @return The name of the state
		 */
		@Override
		public String toString() {
			return this.name;
		}
	} 
	
	/** Material used to highlight the selection. */
	public static final PhongMaterial SELECTED_MATERIAL;
	static {
		SELECTED_MATERIAL = new PhongMaterial(Color.color(0.2, 0.4, 0.6, 0.25));
		SELECTED_MATERIAL.setSpecularColor(Color.color(0, 0, 0, 0));
		SELECTED_MATERIAL.setSpecularPower(250);
	}
	
	/** name of the model. */
	private String name;
	
	/** Path of the directory containing the XML and STL files for this model. */
	private String directoryPath;
	
	/** 
	 * Map associating the different Levels Of Detail (LOD) and the directories containing the geometries of the 
	 * components. 
	 */
	private HashMap<Integer, String> geometriesDirectories;
	
	/** Root component of the model. */
	private Component root;
	
	private State state;
	
	/** Selected components in this model. */
	private CopyOnWriteArraySet<Component> selection;
	
	/** Activation of the axes visualization. */
	private boolean visualizeAxes;

	/**
	 * Default constructor.
	 */
	public Model() {
		this.name = "";
		this.directoryPath = "";
		this.geometriesDirectories = new HashMap<>();
		this.root = null;
		this.selection = new CopyOnWriteArraySet<>();
		this.visualizeAxes = true;
		this.state = State.EXPORTED;
	}
	
	/**
	 * Gets the name of the model.
	 * @return the name
	 */
	public String getName() {
		return this.name;
	}
	
	/**
	 * Gets the directory path of the model
	 * @return The directory path
	 */
	public String getDirectoryPath() {
		return directoryPath;
	}
	
	/**
	 * Gets the state of the model
	 * @return The state
	 */
	public State getState() {
		return state;
	}
	
	/**
	 * The the selected components.
	 * @return The selection
	 */
	public CopyOnWriteArraySet<Component> getSelection() {
		return this.selection;
	}
	
	/**
	 * gets the root component of the model.
	 * @return The root component
	 */
	public Component getRoot() {
		return root;
	}
	
	/**
	 * Gets if the axes visualization is activated.
	 * @return true if the axes visualization is activated
	 */
	public boolean getVisualizeAxes() {
		return this.visualizeAxes;
	}
	
	/**
	 * Sets the model's name.
	 * @param name New name
	 */
	public void setName(String name) {
		this.name = name;
	}
	
	/**
	 * Sets the directory path of the model
	 * @param directoryPath New directory path
	 */
	public void setDirectoryPath(String directoryPath) {
		this.directoryPath = directoryPath;
	}
	
	/**
	 * Gets the map of the geometry directories and their LOD.
	 * @return The geometry directories
	 */
	public HashMap<Integer, String> getGeometriesDirectories() {
		return geometriesDirectories;
	}
	
	/**
	 * Gets a geometry directory from its LOD. 
	 * @param lod Level of detail
	 * @return The directory if the LOD is defined, null otherwise
	 */
	public String getGeometryDirectory(int lod) {
		return this.geometriesDirectories.get(lod);
	}
	
	/**
	 * Sets the root component of the model
	 * @param root New root component
	 */
	public void setRoot(Component root) {
		this.root = root;
	}
	
	/**
	 * Sets the state of the model
	 * @param state New state
	 */
	public void setState(State state) {
		this.state = state;
	}
	
	/**
	 * Activate/deactivate the visualization of the axes.
	 * @param visualizeAxes true to activate
	 */
	public void setVisualizeAxes(boolean visualizeAxes) {
		Logger.getLogger("nomad-3d").info("Setting axes visualization " + ((visualizeAxes) ? "ON" : "OFF"));
		this.visualizeAxes = visualizeAxes;
		updateVisualAxes();
	}
	
	/**
	 * Adds a geometry directory to the model
	 * @param lod Level of detail of the geometries
	 * @param dir Directory path
	 * @return The previous value associated with the LOD, or null if there was directory for the LOD
	 */
	public String addGeometryDirectory(Integer lod, String dir) {
		return this.geometriesDirectories.put(lod, dir);
	}
	
	/**
	 * Adds a component to the selection.
	 * @param c Component to add
	 * @return true if the component was not in the selection already
	 */
	public boolean addToSelection(Component c) {
		Logger.getLogger("nomad-3d").fine("Adding component \"" + c.getName() + "\" to selection...");
		if (c.isLeaf()) {
			MeshView meshView = (MeshView)c.getSceneNode();
			meshView.setMaterial(SELECTED_MATERIAL);
		}
		if (c.getTreeItem() != null) {
			c.getTreeItem().selectedProperty().set(true);
		}
		c.visibleAxis(this.visualizeAxes);
		boolean res = this.selection.add(c);
		Logger.getLogger("nomad-3d").fine("Component added.");
		return res;
	}

	/**
	 * Remove a component from the selection.
	 * @param c Component to remove
	 * @return true if the selection contained the component
	 */
	public boolean removeFromSelection(Component c) {
		return removeFromSelection(c, true);
	}
	
	/**
	 * Remove a component from the selection.
	 * @param c Component to remove
	 * @param updateTreeView true to uncheck the tree item of the component
	 * @return true if the selection contained the component
	 */
	private boolean removeFromSelection(Component c, boolean updateTreeView) {
		Logger.getLogger("nomad-3d").fine("Removing component \"" + c.getName() + "\" from selection...");
		if (c.isLeaf()) {
			MeshView meshView = (MeshView)c.getSceneNode();
			meshView.setMaterial(c.getMaterial());
		}
		if (c.getTreeItem() != null && updateTreeView) {
			c.getTreeItem().selectedProperty().set(false);
		}
		c.visibleAxis(false);
		boolean res = this.selection.remove(c);
		
		// Remove parent for consistency
		if (res && !c.isRoot()) {
			removeFromSelection(c.getParent(), false);
		}
		
		Logger.getLogger("nomad-3d").fine("Component removed.");
		return res;
	}
	
	/**
	 * Clears the selection.
	 */
	public void clearSelection() {
		Logger.getLogger("nomad-3d").info("Clearing selection of the model \"" + this.name + "\"...");
		for (Component comp : this.selection) {
			removeFromSelection(comp);
		}
		Logger.getLogger("nomad-3d").info("Selection cleared.");
	}
	
	/**
	 * Selects all the components of the model.
	 */
	public void selectAll() {
		Logger.getLogger("nomad-3d").info("Selecting all components of the model \"" + this.name + "\"...");
		this.root.setSelected(true, this, true);
		Logger.getLogger("nomad-3d").info("All components selected.");
	}
	
	/**
	 * Clears the model.
	 */
	public void clear() {
		Logger.getLogger("nomad-3d").info("Clearing model \"" + this.name + "\"...");
		this.root = null;
		this.name = "";
		this.directoryPath = "";
		this.selection.clear();
		this.visualizeAxes = true;
		Logger.getLogger("nomad-3d").info("Model cleared.");
	}
	
	/**
	 * Computes the scene hierarchies for all the configurations and the finest LOD.
	 */
	public void computeSceneHierarchies() {
		computeSceneHierarchies(0);
	}
	
	/**
	 * Computes the scene hierarchies for all the configurations.
	 * @param lod Level of detail of the components geometries
	 */
	public void computeSceneHierarchies(int lod) {
		Logger.getLogger("nomad-3d").info("Computing tree view of the model \"" + this.name + "\"...");
		this.root.toTreeItem(this); // must be done before computing scene hierarchy
		Logger.getLogger("nomad-3d").info("Tree view computed.");
		Logger.getLogger("nomad-3d").info("Computing 3D scene graph of the model \"" + this.name + "\"...");
		this.root.computeSceneHierarchy(this, lod);
		Logger.getLogger("nomad-3d").info("Scene graph computed.");
	}
	
	/**
	 * Computes the axes of the model.
	 */
	public void computeAxes() {
		Logger.getLogger("nomad-3d").info("Computing axes of the model \"" + this.name + "\"...");
		this.selection.clear();
		this.root.computeAxes();
		Logger.getLogger("nomad-3d").info("Axes computed.");
		
		if (this.state != State.EXPORTED) {
			computeSceneHierarchies();
		}
	}

	/**
	 * Analyzes the configurations of the model.
	 * @return true if the analysis is successful, false in case of errors
	 */
	public boolean analyzeConfigurations() {
		Logger.getLogger("nomad-3d").info("Analyzing configurations of the model \"" + this.name + "\"...");
		boolean res = this.root.configurationsAnalysis();
		Logger.getLogger("nomad-3d").info("Configurations analyzed.");
		return res;
	}
	
	/**
	 * Shows a configuration
	 * @param configName Name of the configuration
	 */
	public void showConfiguration(String configName) {
		Logger.getLogger("nomad-3d").info("Showing configuration \"" + configName + "\" of the model \"" + this.name + "\"...");
		this.root.resetAxisTransform(true);
		this.root.showConfiguration(configName);
		Logger.getLogger("nomad-3d").info("Configuration shown.");
	}
	
	/**
	 * Resets the movement of every component along its axis.
	 */
	public void resetMovement() {
		Logger.getLogger("nomad-3d").info("Resetting movement of the model \"" + this.name + "\"...");
		this.root.resetAxisTransform(true);
		Logger.getLogger("nomad-3d").info("Movement reset.");
	}
	
	/**
	 * Sets the movement of every component along its axis to the zero position.
	 */
	public void setToZeroAxisTransform() {
		Logger.getLogger("nomad-3d").info("Setting movement of the model \"" + this.name + "\" to zero...");
		this.root.setToZeroAxisTransform(true);
		Logger.getLogger("nomad-3d").info("Movement set to zero.");
	}
	
	@Override
	public String toString() {
		String str = "Model [\n";
		str += "Name : " + name + "\n";
		str += "Root : " + root + "\n";
		str += "Visualize axes : " + visualizeAxes + "\n";
		str += "]";
		return str;
	}
	
	/**
	 * Updates the visual axes for all configurations.
	 */
	private void updateVisualAxes() {
		Logger.getLogger("nomad-3d").info("Updating visual axes of the model \"" + this.name + "\"...");
		this.root.updateVisualAxes(this.selection, this.visualizeAxes);
		Logger.getLogger("nomad-3d").info("Visual axes updated.");
	}
}
